(*
     Catcher! OCR Module  -  Port of JAntiCaptcha
     Copyright (C) 2010  CandiSoft

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

unit AntiCaptcha;

interface

uses Windows, SysUtils, Classes, Graphics, CSTools, XMLDoc, XMLIntf, Generics.Collections,
     JACScript, untLogger, Captcha, LetterComperator, Letter, Math, Utilities,
     PNGImage, GIFImg, Jpeg, Ole2;

type
  TAntiCaptcha = class
  private
    FMethodPath: string;
    FWorkingCaptcha: TCaptcha;
    FSourceImage: TGraphic;
    FLetterNum, FMinLetterNum, FMaxLetterNum: Integer;
    FImageType: string;
    FLetterDB: TList<TLetter>;
    FScript: TJACScript;
    FExtern: Boolean;
    FExtCmd, FExtSrc, FExtDst: string;

    procedure GetJACInfo;
    procedure CreateLetterDBFromMTH(Mth: IXMLDocument);

    function CallExtern: string;

    procedure SetImageType(const Value: string);
    procedure SetLetterNum(const Value: Integer);
    procedure SetMinLetterNum(const Value: Integer);
    procedure SetMaxLetterNum(const Value: Integer);
  public
    class function GetCaptchaCode(Img: TBitmap; const MethodPath: string): string;

    constructor Create(const AMethodPath: string);
    destructor Destroy; override;

    function GetLetter(Letter: TLetter): TLetterComperator;
    function GetLetterExtended(Letter: TLetter): TLetterComperator;
    function CheckCaptcha(Captcha: TCaptcha): string; overload;
    function CheckCaptcha(CaptchaFile: string): string; overload;
    function CreateCaptcha(CaptchaImage: TBitmap): TCaptcha;

    procedure LoadMTHFile(FileName: string);

    property ImageType: string read FImageType write SetImageType;
    property LetterNum: Integer read FLetterNum write SetLetterNum;
    property MinLetterNum: Integer read FMinLetterNum write SetMinLetterNum;
    property MaxLetterNum: Integer read FMaxLetterNum write SetMaxLetterNum;
    property MethodPath: string read FMethodPath;
    property Script: TJACScript read FScript;
  end;

  TSpecial = class
  public
    class function GetLetters(Captcha: TCaptcha): TArrayofLetter; virtual; abstract;
  end;
  TSpecialClass = class of TSpecial;

var
  SpecialClasses: TDictionary<string, TSpecialClass>;

implementation

class function TAntiCaptcha.GetCaptchaCode(Img: TBitmap; const MethodPath: string): string;
var
  JAC: TAntiCaptcha;
  Cap: TCaptcha;
begin
  JAC := TAntiCaptcha.Create(MethodPath);
  try
    Cap := JAC.CreateCaptcha(img);
    if Cap = nil then
    begin
      Logger.Severe('Captcha Bild konnte nicht eingelesen werden!');
      Exit('ERROR');
    end;

    // frmGUI.ShowImage(Cap.GetImageWithGaps(2));
    Result := JAC.CheckCaptcha(Cap);
    Logger.Info('Captcha Text: ' + Result);
  finally
    JAC.Free;
  end;
end;

constructor TAntiCaptcha.Create(const AMethodPath: string);
begin
  FLetterDB := nil;

  FMethodPath := IncludeTrailingPathDelimiter(AMethodPath);
  Logger.Info(FMethodPath);

  GetJACInfo;

  FScript := TJACScript.Create(Self, FMethodPath);

  LoadMTHFile('letters.mth');

  if Assigned(FLetterDB) then
    Logger.Fine('Letter DB loaded: Buchstaben: ' + IntToStr(FLetterDB.Count));
end;

destructor TAntiCaptcha.Destroy;
begin
  CoUninitialize;

  FScript.Free;

  if Assigned(FLetterDB) then
    FLetterDB.Free;

  inherited;
end;

procedure TAntiCaptcha.GetJACInfo;
var
  Doc: IXMLDocument;
  NL: IXMLNodeList;
  i: Integer;
begin
  if not FileExists(FMethodPath + 'jacinfo.xml') then
  begin
    Logger.Severe('jacinfo.xml is missing!');
    Exit;
  end;

  CoInitialize(nil);

  Doc := NewXMLDocument;
  Doc.LoadFromFile(FMethodPath + 'jacinfo.xml');
  Doc.Active := True;

  NL := Doc.DocumentElement.ChildNodes;

  for i := 0 to NL.Count - 1 do
    if NL[i].NodeName = 'method' then
    begin
      if NL[i].HasAttribute('type') then
        FExtern := SameText(NL[i].Attributes['type'], 'extern')
      else
        FExtern := False;
    end
    else if NL[i].NodeName = 'format' then
    begin
      if NL[i].HasAttribute('letterNum') then
        FLetterNum := NL[i].Attributes['letterNum'];

      if NL[i].HasAttribute('minLetterNum') then
        FMinLetterNum := NL[i].Attributes['minLetterNum'];

      if NL[i].HasAttribute('maxLetterNum') then
        FMaxLetterNum := NL[i].Attributes['MaxLetterNum'];

      ImageType := NL[i].Attributes['type'];
    end
    else if NL[i].NodeName = 'command' then
    begin
      FExtCmd := StringReplace(NL[i].Attributes['cmd'], 'jd\captcha', 'ocr', [rfReplaceAll]);
      FExtSrc := StringReplace(NL[i].Attributes['src'], 'jd\captcha', 'ocr', [rfReplaceAll]);
      FExtDst := StringReplace(NL[i].Attributes['dst'], 'jd\captcha', 'ocr', [rfReplaceAll]);
    end;
end;

procedure TAntiCaptcha.LoadMTHFile(FileName: string);
var
  Mth: IXMLDocument;
begin
  if not FileExists(FMethodPath + FileName) then
  begin
    Logger.Severe(FileName + ' is missing!');
    Exit;
  end;

  Mth := NewXMLDocument;
  Mth.LoadFromFile(FMethodPath + FileName);
  Mth.Active := True;

  CreateLetterDBFromMTH(Mth);
end;

procedure TAntiCaptcha.CreateLetterDBFromMTH(Mth: IXMLDocument);
var
  Tmp: TLetter;
  NL: IXMLNodeList;
  i: Integer;
  ID: string;
begin
  if Assigned(FLetterDB) then
  begin
    for Tmp in FLetterDB do
      Tmp.Free;

    FLetterDB.Free;
  end;

  FLetterDB := TList<TLetter>.Create;

  try
    if Mth = nil then
      Exit;

    NL := Mth.DocumentElement.ChildNodes;

    for i := 0 to NL.Count - 1 do
    begin
      if NL[i].NodeName = 'letter' then
      begin
        Tmp := TLetter.Create;
        Tmp.Owner := Self;

        if NL[i].HasAttribute('id') then
          ID := NL[i].Attributes['id']
        else
          ID := '';

        if not Tmp.SetTextGrid(NL[i].Text) then
        begin
          Logger.Severe('Error in Letters DB line: ' + IntToStr(i) + ':' + NL[i].Text + ' ID:' + ID);
          Continue;
        end;

        if ID <> '' then
          Tmp.ID := StrToInt(ID);

        Tmp.SourceHash := NL[i].Attributes['captchaHash'];
        Tmp.DecodedValue := NL[i].Attributes['value'];
        Tmp.BadDetections := StrToInt(NL[i].Attributes['bad']);
        Tmp.GoodDetections := StrToInt(NL[i].Attributes['good']);
        FLetterDB.Add(Tmp);
      end
      else
        Logger.Warning('UNIMPLEMENTED (MTH): ' + NL[i].NodeName);
    end;
  except
    Logger.Severe('Fehler beim Lesen der MTH Datei! Methode kann nicht funktionieren!'#13#10 +
                  '(E: ' + Exception(ExceptObject).Message);
  end;
end;

function TAntiCaptcha.GetLetter(Letter: TLetter): TLetterComperator;
var
  Res, LC: TLetterComperator;
  LastPercent: Double;
  BvX, BvY, i: Integer;
  Tmp: TLetter;
begin
  Result := nil;

  if Double(FScript.GetParam('quickScanValityLimit')) <= 0 then
  begin
    Logger.Info('quickscan disabled');
    Exit;
  end;

  Logger.Info('Work on Letter: ' + Letter.ToString);
  Res := nil;
  LastPercent := 100;

  try
    if FLetterDB = nil then
    begin
      Logger.Severe('LetterDB nicht vorhanden!');
      Exit;
    end;
    BvX := FScript.GetParam('borderVarianceX');
    BvY := FScript.GetParam('borderVarianceY');

    LC := TLetterComperator.Create(Letter, nil);
    LC.SetScanVariance(0, 0);
    LC.Owner := Self;
    Res := LC;

    Logger.Info('Do quickscan');

    // xxx special handling

    for i := 0 to FLetterDB.Count - 1 do
    begin
      Tmp := FLetterDB[i];
      if (Abs(Tmp.Height - Letter.Height) > BvY) or (Abs(Tmp.Width - Letter.Width) > BvX) then
        Continue;

      LC := TLetterComperator.Create(Letter, Tmp);
      LC.Owner := Self;

      // xxx special handling

      LC.Run;

      if (Res = nil) or (LC.ValityPercent < Res.ValityPercent) then
      begin
        if (Res <> nil) and (Res.ValityPercent < LastPercent) then
          LastPercent := Res.ValityPercent;
        Res := LC;

        if Double(FScript.GetParam('LetterSearchLimitPerfectPercent')) >= LC.ValityPercent then
        begin
          Logger.Finer('Perfect Match: ' + Res.B.DecodedValue + ' ' +
                        FloatToStr(Res.ValityPercent) +
                        ' good:' + IntToStr(Res.B.GoodDetections) +
                        ' bad: ' + IntToStr(Res.B.BadDetections));

          Res.DetectionType := QUICKSCAN_PERFECT_MATCH;
          Res.Reliability := LastPercent - Res.ValityPercent;
          Exit(Res);
        end;
      end
      else
      if Res <> nil then
      begin
        if LC.ValityPercent < LastPercent then
          LastPercent := LC.ValityPercent;

        LC.Free;
      end;
    end;
  except
    Logger.Severe('GetLetter: ' + Exception(ExceptObject).Message);
  end;

  if (Res <> nil) and (Res.B <> nil) then
  begin
    Logger.Finer('Normal Match: ' + Res.B.DecodedValue + ' ' +
                  FloatToStr(Res.ValityPercent) +
                  ' good:' + IntToStr(Res.B.GoodDetections) +
                  ' bad: ' + IntToStr(Res.B.BadDetections));

    Res.Reliability := LastPercent - Res.ValityPercent;

    if (Res.Reliability >= Double(FScript.GetParam('quickScanReliabilityLimit'))) and
       (Res.ValityPercent < Double(FScript.GetParam('quickScanValityLimit'))) then
    begin
      Res.DetectionType := QUICKSCAN_MATCH;
      Result := Res;
    end
    else
    begin
      Logger.Warning(Format('Letter nicht ausreichend erkannt. Try Extended %f - %f | %f - %f',
                            [Res.Reliability, Double(FScript.GetParam('quickScanReliabilityLimit')),
                             Res.ValityPercent, Double(FScript.GetParam('quickScanValityLimit'))]));
      Result := GetLetterExtended(Letter);
    end;
  end
  else
  begin
    Logger.Warning('Letter nicht erkannt. Try Extended');
    Result := GetLetterExtended(Letter);
  end;
end;

function TAntiCaptcha.GetLetterExtended(Letter: TLetter): TLetterComperator;
var
  Res, LC: TLetterComperator;
  LastPercent: Double;
  Tmp, OrgLetter, Ltr: TLetter;
  LeftAngle, RightAngle, Temp, Steps, Angle: Integer;
  TurnDB: Boolean;
begin
  Res := nil;
  Logger.Info('Extended Scan');
  LastPercent := 100;
  try
    if FLetterDB = nil then
    begin
      Logger.Severe('LetterDB nicht vorhanden!');
      Exit(nil);
    end;

    LeftAngle := FScript.GetParam('scanAngleLeft');
    RightAngle := FScript.GetParam('scanAngleRight');
    if LeftAngle > RightAngle then
    begin
      Temp := LeftAngle;
      LeftAngle := RightAngle;
      RightAngle := Temp;
      Logger.Warning('param.scanAngleLeft > param.scanAngleRight');
    end;
    Steps := Max(1, Integer(FScript.GetParam('scanAngleSteps')));
    TurnDB := FScript.GetParam('turnDB');

    OrgLetter := Letter;
    LC := TLetterComperator.Create(Letter, nil);
    LC.Owner := Self;
    Res := LC;

    Angle := GetJumperStart(LeftAngle, RightAngle);
    while CheckJumper(Angle, LeftAngle, RightAngle) do
    begin
      if TurnDB then
        Letter := OrgLetter
      else
        Letter := OrgLetter.Turn(Angle);

      for Ltr in FLetterDB do
      begin
        if TurnDB then
          Tmp := Ltr.Turn(Angle)
        else
          Tmp := Ltr;

        if (Abs(Tmp.Height - Letter.Height) > Integer(FScript.GetParam('borderVarianceY'))) or
           (Abs(Tmp.Width - Letter.Width) > Integer(FScript.GetParam('borderVarianceX'))) then
          Continue;

        LC := TLetterComperator.Create(Letter, Tmp);
        LC.Owner := Self;
        LC.Run;

        if (Res = nil) or (LC.ValityPercent < Res.ValityPercent) then
        begin
          if (Res <> nil) and (Res.ValityPercent < LastPercent) then
            LastPercent := Res.ValityPercent;
          Res := LC;

          if Double(FScript.GetParam('LetterSearchLimitPerfectPercent')) >= LC.ValityPercent then
          begin
            Logger.Finer('Perfect Match: ' + Res.B.DecodedValue + ' ' +
                          FloatToStr(Res.ValityPercent) +
                          ' good:' + IntToStr(Res.B.GoodDetections) +
                          ' bad: ' + IntToStr(Res.B.BadDetections));

            Res.DetectionType := PERFECT_MATCH;
            Res.Reliability := LastPercent - Res.ValityPercent;
            Exit(Res);
          end;
        end
        else
        if Res <> nil then
        begin
          if LC.ValityPercent < LastPercent then
            LastPercent := LC.ValityPercent;

          LC.Free;
        end;
      end;

      Angle := NextJump(Angle, LeftAngle, RightAngle, Steps);
    end;
  except
    Logger.Severe('GetLetterExtended: ' + Exception(ExceptObject).Message);
  end;

  if (Res <> nil) and (Res.B <> nil) then
  begin
    Logger.Finer('Normal Match: ' + Res.B.DecodedValue + ' ' +
                  FloatToStr(Res.ValityPercent) +
                  ' good:' + IntToStr(Res.B.GoodDetections) +
                  ' bad: ' + IntToStr(Res.B.BadDetections));

    Res.Reliability := LastPercent - Res.ValityPercent;
  end
  else
    Logger.Severe('Letter endgltig nicht erkannt.');

  Result := Res;
end;

function TAntiCaptcha.CheckCaptcha(Captcha: TCaptcha): string;
var
  Letters: TArrayofLetter;
  Correct: Double;
  Akt: TLetterComperator;
  NewLettersList: TList<TLetterComperator>;
  i: Integer;
begin
  if FExtern then
    Exit(CallExtern());

  FWorkingCaptcha := Captcha;

  // Captcha vorbereiten
  FScript.ExecutePrepareCommands(Captcha);

  Letters := Captcha.GetLetters(LetterNum);
  if Letters = nil then
  begin
    Captcha.ValityPercent := 100;
    Logger.Severe('Captcha konnte nicht erkannt werden!');
    Exit('');
  end;

  Result := '';
  Correct := 0;

  NewLettersList := TList<TLetterComperator>.Create;
  for i := 0 to High(Letters) do
  begin
    Letters[i].ID := i;
    if Letters[i].Detected <> nil then
      Akt := TLetterComperator(Letters[i].Detected)
    else
      Akt := GetLetter(Letters[i]);
    Akt.A.ID := i;

    NewLettersList.Add(Akt);
  end;

  for Akt in NewLettersList do
  begin
    if (Akt = nil) or (Akt.ValityPercent >= 100) then
    begin
      Result := Result + '-';
      Correct := Correct + 100;
    end
    else
    begin
      Result := Result + Akt.GetDecodedValue;
      Correct := Correct + Akt.ValityPercent;
    end;
  end;

  if NewLettersList.Count = 0 then
  begin
    Captcha.ValityPercent := 100;
    Exit('');
  end;

  Captcha.ValityPercent := Correct / NewLettersList.Count;
end;

procedure ExternWait(const ProcessInfo: TProcessInformation;
  var ATerminate: Boolean; var MaxWait: Integer);
begin
  if MaxWait <= 0 then
  begin
    ATerminate := True;
    Exit;
  end;

  Sleep(1000);
  Dec(MaxWait, 1000);
end;

function TAntiCaptcha.CallExtern: string;
begin
  FSourceImage.SaveToFile(ExtractFilePath(ParamStr(0)) + FExtSrc);
  ExecuteFile(ExtractFilePath(ParamStr(0)) + FExtCmd, SW_SHOW, '', '', True, ExternWait, 120000);
  Result := Trim(LoadStringFromFile(ExtractFilePath(ParamStr(0)) + FExtDst));
end;

function TAntiCaptcha.CheckCaptcha(CaptchaFile: string): string;
var
  BMP: TBitmap;
  Cap: TCaptcha;
  i: Integer;
begin
  for i := 0 to 2 do
  begin
    try
      case i of
        0: FSourceImage := TGIFImage.Create;
        1: FSourceImage := TPNGImage.Create;
        else FSourceImage := TJPEGImage.Create;
      end;

      FSourceImage.LoadFromFile(CaptchaFile);
      Break;
    except
    end;

    // Nothing found
    if i = 2 then
    begin
      Logger.Severe('Unknown image format: ' + CaptchaFile);
      Exit;
    end;
  end;

  BMP := TBitmap.Create;
  BMP.Assign(FSourceImage);

  Cap := CreateCaptcha(BMP);
  try
    Result := CheckCaptcha(Cap);

    if Assigned(Cap) then
      Logger.Log(Format('Recognition done - Vality: %f; Code: %s', [Cap.ValityPercent, Result]))
    else if FExtern then
      Logger.Log(Format('External recognition done - Code: %s', [Result]))
  finally
    FSourceImage.Free;
    BMP.Free;
    if Assigned(Cap) then
      Cap.Free;
  end;
end;

function TAntiCaptcha.CreateCaptcha(CaptchaImage: TBitmap): TCaptcha;
begin
  if FExtern then
    Exit(nil);

  Result := TCaptcha.GetCaptcha(CaptchaImage, Self);
  Result.Owner := Self;
end;

procedure TAntiCaptcha.SetImageType(const Value: string);
begin
  Logger.Finer('SET PARAMETER: [imageType] = ' + Value);

  FImageType := Value;
end;

procedure TAntiCaptcha.SetLetterNum(const Value: Integer);
begin
  Logger.Finer('SET PARAMETER: [letterNum] = ' + IntToStr(Value));

  FLetterNum := Value;
end;

procedure TAntiCaptcha.SetMaxLetterNum(const Value: Integer);
begin
  // TODO?
end;

procedure TAntiCaptcha.SetMinLetterNum(const Value: Integer);
begin
  // TODO?
end;

initialization
  SpecialClasses := TDictionary<string, TSpecialClass>.Create;

finalization
  SpecialClasses.Free;

end.
