(*
     Catcher! OCR Module  -  Port of JAntiCaptcha
     Copyright (C) 2010  CandiSoft

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

unit JACScript;

interface

uses Variants, SysUtils, CSTools, PixelGrid, UTILITIES, Captcha,
     Letter, untLogger, Generics.Collections;

type
  TCCArray = array[0..2] of Integer;

  TJACScript = class
  private
    FCaptchaPrepareCommands, FJacCommands, FLetterCommands: TList<TArrayofstring>;
    FGaps: TArrayofInteger;
    FColorComponents: TCCArray;  // = [3, 3, 3]
    FColorFaktor, FColor: Integer;
    FOwner: TObject;
    FScriptFile, FMethod: string;
    FParameters: TDictionary<string, Variant>;

    procedure Init;
    procedure ExecuteParameterCommands;
    procedure ParseScriptFile;
    function ParseCommand(Cmd: string): TArrayofstring;
    function ToType(Arg, Key: string): Variant;
  public
    constructor Create(Owner: TObject; Method: string);
    destructor Destroy; override;

    procedure SetParam(Key: string; Value: Variant);
    function GetParam(Key: string): Variant;
    procedure ExecutePrepareCommands(Captcha: TCaptcha);
    procedure ExecuteLetterPrepareCommands(Letter: TLetter);

    procedure SetColorType(CType: string);

    property ColorComponents: TCCArray read FColorComponents;
    property ColorFaktor: Integer read FColorFaktor;
    property ColorFormat: Integer read FColor;
    property Gaps: TArrayofInteger read FGaps;
  end;

implementation

uses AntiCaptcha;

constructor TJACScript.Create(owner: TObject; method: string);
begin
  DecimalSeparator := '.';

  FParameters := TDictionary<string, Variant>.Create;
  Init;
  FOwner := Owner;
  FMethod := Method;
  FScriptFile := 'script.jas';
  ParseScriptFile;
  ExecuteParameterCommands;
end;

destructor TJACScript.Destroy;
begin
  FParameters.Free;

  inherited;
end;

procedure TJACScript.Init;
begin
  // Prozentwert. Ab dieser Schwelle an Korektheit wird ein Letter als
  // 100% richtig gewertet
  SetParam('LetterSearchLimitPerfectPercent', 10.0);

  // Kontrastwert fr die Erkennung ob ein Pixel Farblich zu einem objekt
  // passt (Kontrast Objektdurchschnitt/pixel)
  SetParam('objectColorContrast', 0.3);

  // Kontrastwert zur erkennung eines ObjektPixels (Kontrast Objekt/hintergrund)
  SetParam('objectDetectionContrast', 0.5);

  // Minimale Objektflche fr die Objekterkennung.
  SetParam('minimumObjectArea', 200);

  // Gibt an ob die langsamere Object Detection anstelle der sonst
  // blichen Reihen Detection verwendet werden soll.
  SetParam('useobjectDetection', False);
  SetParam('useColorObjectDetection', False);
  SetParam('colorObjectDetectionPercent', 15);
  SetParam('colorObjectDetectionRunningAverage', 255);

  // Parameter: Gibt die Anzahl der Reihen(Pixel) an die zur peak
  // detection verwendet werden sollen.
  SetParam('gapWidthPeak', 1);

  // Gibt die Verwendete Farbkombination an. Siehe JAC Script Docu fr
  // genauere Infos.
  SetColorType('h');

  // Parameter: Gibt die Anzahl der reihen an die zur Average Detection
  // verwendet werden sollen
  SetParam('gapWidthAverage', 1);

  { Parameter: gapAndAverageLogic=true: Es werden Lcken verwendet bei
    denen Peak und Average detection zusammenfallen (AND)
    gapAndAverageLogic=false: Es werden sowohl peak als Auch Average
    Lcken verwendet (nur in Ausnahmefllen) (OR)
  }
  SetParam('gapAndAverageLogic', True);

  // Parameter: Der Kontrastwert fr die Average Detection. ~1
  SetParam('gapDetectionAverageContrast', 1.3);

  // Parameter: Der Kontrastwert fr die Peak Detection. ~0.25
  SetParam('gapDetectionPeakContrast', 0.25);

  SetParam('useSpecialGetLetters', '');
  SetParam('useLetterFilter', '');
  SetParam('useLettercomparatorFilter', '');
  SetParam('preValueFilter','');
  SetParam('postValueFilter','');
  SetParam('comparatorExtension','');

  SetParam('LetterSearchLimitFalsePercent', 100);

  // Parameter: Average Detection verwenden
  SetParam('useAverageGapDetection', false);

  // Winkel Schrittgre bei der automatischen Ausrichtung
  SetParam('alignAngleSteps', 5);

  SetParam('splitGapsOverlap', 4);

  // Linker Winkelrand beim Scannen
  SetParam('scanAngleLeft', 0);
  // Rechter Winkelrand beim Scannen
  SetParam('scanAngleRight', 0);
  // Winkelschritte beim Scannen
  SetParam('scanAngleSteps', 5);

  // Parameter: Peak Detection verwenden
  SetParam('usePeakGapdetection', True);

  // Parameter: Kontrollwert ber die minimale Buchstabenbreite
  SetParam('minimumLetterWidth', 0);

  (*
   * Parameter: Wert gibt an um welchen faktor die Fingerprints
   * verkleinert werden. So gro wie mglich, so klein wie ntig. Wenn
   * dieser Wert verndert wird, wrd die MTH File unbrauchbar und muss neu
   * trainiert werden
   *)
  SetParam('simplifyFaktor', 1.0);

  // A: Captchabild | B: Datenbankbild
  // Gewichtung des A Fehlers
  SetParam('errorAWeight', 1.0);
  // Gewichtung des B fehlers
  SetParam('errorBWeight', 1.0);

  // Bricht die Erkennung ab wenn die Object Erkennung fehlgeschlagen ist
  SetParam('cancelIfObjectDetectionFailed', false);

  (*
   * Gewichtung des grenunterschieds Ausschnitt/Originale(Datenbank) Ist
   * ein Datenbank-Buchstabe nur teilweise auf dem Trefferausschnitt,
   * verschlechtert das die Wertung
   *)
  SetParam('intersectionDimensionWeight', 0.9);

  // Gibt an wie viele berlagerungsstrungen ausgefiltert werden sollen
  // -1: >keine 0: bei 0 Nachbarn 1: bei hchstem einen Nachbarn etc
  SetParam('overlayNoiseSize', -1);


  // Gewichtung der A-Abdeckung
  SetParam('coverageFaktorAWeight', 1.0);
  // Gewichtung der B-Abdeckung
  SetParam('coverageFaktorBWeight', 1.0);

  // Reliabilitywert ab dem ein Quickscan gltig ist
  SetParam('quickScanReliabilityLimit', 2.0);

  // ValityPercentWert ab dem ein quickscan angenommen wird (perfectmatch)
  SetParam('quickScanValityLimit', 50.0);

  // Minimale Objektgre die als Zerklftung gezhlt wird
  SetParam('minCleftSize', 3);
  // Gewichtung der Schittmengenzerkftuntung
  SetParam('cleftFaktor', 0.3);

  // Schwellwert fr den Quickscanfilter
  SetParam('preScanFilter', 100);

  (*
   * Schwellwert fr den quickscanfilter. Wird dieser emergeny Wert
   * gesetzt, so wird bei Erfolgloser Suche der filter auf diesen wert
   * gesetzt und ein 2. Lauf gestartet
   *)
  SetParam('preScanEmergencyFilter', 0);
  // Anzahl der Linien der prescan auswerten soll
  SetParam('preScanFaktor', 3);
  // Anzahld er Pixel die beim Trennen von Objekten doppelt gezhlt werden (berlappung)
  SetParam('splitPixelObjectsOverlap', 4);

  // Parameter: Allgemeiner Bildkontrastparameter ~0.8 bis 1.2
  SetParam('relativeContrast', 0.85);

  // Parameter: Gibt die Tolleranz beim Subern des Hintergrunds an ~0.05-0.5
  SetParam('backgroundSampleCleanContrast', 0.1);

  // Parameter: Gibt fr dieverse SW Umwandlungen den Schwellwert an
  SetParam('blackPercent', 0.1);

  // Ob bei EasyCaptcha Buchstaben als Schwarzwei gewertet werden sollen
  SetParam('easyCaptchaBW', False);

  // Gibt an ob beim Training nur falscherkannte Buchstaben gespeichert
  // werden (true) oder alle (False)
  SetParam('trainOnlyUnknown', True);

  // Gibt an ob bei der objekterenntung pixeln gefolgt wird die quer liegen.
  SetParam('followXLines',true);
  SetParam('turnDB',false);

  (*
   * Parameter: Scan-Parameter. Gibt an um wieviele Pixel sich Letter und
   * Vergleichsletter unterscheiden drfen um verglichen zu werden. Hohe
   * Werte machen das ganze Langsam
   *)
  SetParam('borderVarianceX', 0);
  SetParam('borderVarianceY', 0);

  // Die Zeichenanzahl wird automatisch gesetzt
  SetParam('autoLetterNum', false);

  // Gleich nach der Objekterkennung wird versucht den buchstaben zu
  // erkennung so knnen zusammenhngende Buchstaben miterkannt werden
  SetParam('directLetterDetection', false);

  SetParam('prescandivider',4.0);
  SetParam('divider',6.0);

  (* Parameter: Scan-Parameter. Gibt an um wieviele Pixel Letter und
   * Vergleichsletter gegeneinander verschoben werden um die beste
   * bereinstimung zu finden. Hohe werte verlangemmen die Erkennung
   * deutlich
   *)
  SetParam('scanVarianceX', 0);
  SetParam('scanVarianceY', 0);

  (* BestehenBuchstaben aus mehreren getrennten parts, werdens ie durch
   * die object detection getrennt. Dieser wert gibt fr ein neues objekt
   * an wie weit es sich vom mittelpunkt des letzten objekts entfernt
   * aufhalten darf ohne zum letzten gezhlt zu werden. beispiel: 7.0: der
   * innere kreis von einer 0 wird zum ueren gezhlt wenn sich der
   * innere nicht weiter als 7 % vom mittelunkte des ueren befindet.
   *)
  SetParam('multiplePartMergeMinSize', 0);

  SetParam('abortDirectDetectionOnDetectionError', false);

  (*
    * InverseFontWeight Unterschreitet die Pixelanzahl einer Intersection
    * einen gewissen Teil der Intersectionflche, wird der fehler auf 100%
    * angehoben. Beispiel: inverseFontWeight=8 Die gemeinsammen Pixel einer
    * 400 px Intersection betragen nur 20 pixel. 20*8 = 160; 160<400
    * =>Der Treffer wird nicht gewertet, da die Intersection zu wenig
    * gemeinsamme Pixel hat.
  *)
  SetParam('inverseFontWeight', 8.0);

  SetParam('scanstepx', 1);
  SetParam('scanstepy', 1);
  SetParam('intersectionAHeightWeight', 0.0);
  SetParam('intersectionAWidthWeight', 0.0);
  SetParam('minObjectSize', 10);
end;

procedure TJACScript.SetParam(Key: string; Value: Variant);
begin
  FParameters.AddOrSetValue(LowerCase(Key), Value);
end;

function TJACScript.GetParam(Key: string): Variant;
begin
  if FParameters.ContainsKey(LowerCase(Key)) then
    Result := FParameters[LowerCase(Key)]
  else
    Result := Variants.Null;
end;

procedure TJACScript.ExecuteLetterPrepareCommands(Letter: TLetter);
var
  Params, Cmd: TArrayofstring;
  i: Integer;
  Tmp: TLetter;
begin
  Logger.Fine('Script.jas - Preparing Letter');

  try
    for i := 0 to FLetterCommands.Count - 1 do
    begin
      Cmd := FLetterCommands[i];

      Logger.Info('Executing Function: ' + Cmd[1] + '(' + Cmd[2] + ')');

      if (Cmd[0] = 'parameter') then
      begin
        Logger.Severe('Syntax Error in ' + FMethod + '\script.jas(letter)');
        // captchaPrepareCommands
      end
      else
      if (Cmd[0] = 'function') and (Cmd[2] = '') then
      begin
        if LowerCase(Cmd[1]) = 'normalize' then
        begin
          Letter.Normalize;
          Continue;
        end
        else
        if LowerCase(Cmd[1]) = 'clean' then
        begin
          Letter.Clean;
          Continue;
        end
        else
        if LowerCase(Cmd[1]) = 'toblackandwhite' then
        begin
          Letter.ToBlackAndWhite;
          Continue;
        end
        else
        if LowerCase(Cmd[1]) = 'invert' then
        begin
          Letter.Invert;
          Continue;
        end
        else
          Logger.Severe('Error in ' + FMethod + '\script.jas : Function not valid: ' + cmd[1] + '(' + cmd[2] + ')');
      end
      else
      if (Cmd[0] = 'function') and (StrCount(Cmd[2], ',') = 0) then
      begin
        Params := Explode(',', Cmd[2]);

        if LowerCase(Cmd[1]) = 'toblackandwhite' then
        begin
          Letter.ToBlackAndWhite(StrToFloat(Trim(Params[0])));
          Continue;
        end
        else
        if LowerCase(cmd[1]) = 'resizetoheight' then
        begin
          Letter.ResizeToHeight(StrToInt(Trim(Params[0])));
          Continue;
        end
        else
          Logger.Warning('LETTER COMMAND UNIMPLEMENTED: ' + Cmd[1] + '!');
      end
      else
      if (Cmd[0] = 'function') and (StrCount(Cmd[2], ',') = 1) then
      begin
        Params := Explode(',', Cmd[2]);

        if LowerCase(Cmd[1]) = 'reduceWhiteNoise' then
        begin
          Letter.ReduceWhiteNoise(StrToInt(Trim(Params[0])), StrToFloat(Trim(Params[1])));
          Continue;
        end
        else
        if LowerCase(Cmd[1]) = 'resizetoHeight' then
        begin
          Letter.ResizetoHeight(StrToInt(Trim(Params[0])), StrToFloat(Trim(Params[1])));
          Continue;
        end
        else
        if LowerCase(Cmd[1]) = 'reduceBlackNoise' then
        begin
          Letter.ReduceBlackNoise(StrToInt(Trim(Params[0])), StrToFloat(Trim(Params[1])));
          Continue;
        end
        else
        if LowerCase(Cmd[1]) = 'align' then
        begin
          Tmp := Letter.Align(StrToInt(Trim(Params[0])), StrToInt(Trim(Params[1])));

          Letter.grid := tmp.grid;
          Continue;
        end
        else
          Logger.Warning('LETTER COMMAND UNIMPLEMENTED: ' + Cmd[1] + '!');
      end;

    end;
  except
    Logger.Severe('Syntax Error in ' + FMethod + '\script.jas (exception occured)');
  end;
end;

procedure TJACScript.ExecuteParameterCommands;
var
  i, j: Integer;
  Cmd, Gaps: TArrayofstring;
  NewGaps: TArrayofInteger;
begin
  if (FJacCommands = nil) or (FJacCommands.Count = 0) then
  begin
    Logger.Warning('KEINE JAC COMMANDS');
    Exit;
  end;

  Logger.Fine('Script.jas - Setting script parameters');

  try
    for i := 0 to FJacCommands.Count - 1 do
    begin
      Cmd := FJacCommands[i];

      if Cmd[0] = 'parameter' then
      begin
        if LowerCase(Cmd[1]) = 'gaps' then
        begin
          Cmd[2] := Copy(Cmd[2], 2, Length(Cmd[2]) - 2);
          Gaps := Explode(',', Cmd[2]);
          SetLength(NewGaps, Length(Gaps));
          for j := 0 to High(Gaps) do
            NewGaps[j] := StrToInt(Gaps[j]);

          FGaps := NewGaps;
        end
        else
        if LowerCase(Cmd[1]) = 'colortype' then
          SetColorType(Cmd[2])
        else
        begin
          Logger.Finer(Cmd[1] + ' - ' + Cmd[2]);

          SetParam(LowerCase(Cmd[1]), ToType(Cmd[2], Cmd[1]));
        end;
      end
      else
        Logger.Severe('Syntax Error in ' + FMethod + '\script.jas (' + cmd[0] + ')');
    end;
  except
    Logger.Severe('Syntax Error in ' + FMethod + '\script.jas (exception occured)');
  end;
end;

procedure TJACScript.ExecutePrepareCommands(Captcha: TCaptcha);
var
  Params, Cmd: TArrayofstring;
  i: Integer;
begin
  if (Captcha = nil) or (Captcha.IsPrepared) then    // Ist schon prepared
    Exit;

  Captcha.IsPrepared := True;
  try
    for i := 0 to FCaptchaPrepareCommands.Count - 1 do
    begin
      Cmd := FCaptchaPrepareCommands[i];

      Logger.Fine('Executing Function: ' + Cmd[1] +'(' + Cmd[2] + ')');

      if Cmd[0] = 'parameter' then
        Logger.Severe('Syntaxerror in ' + FMethod + '\script.jas')
      else
        if (Cmd[0] = 'function') and (Cmd[2] = '') then
        begin
          if LowerCase(Cmd[1]) = 'invert' then
          begin
            Captcha.Invert;
            Continue;
          end
          else
          if LowerCase(Cmd[1]) = 'toblackandwhite' then
          begin
            Captcha.ToBlackAndWhite;
            Continue;
          end
          else
          if LowerCase(Cmd[1]) = 'normalize' then
          begin
            Captcha.Normalize;
            Continue;
          end
          else
          if LowerCase(Cmd[1]) = 'clean' then
          begin
            Captcha.Clean;
            Continue;
          end
          else
            Logger.Severe('Error in ' + FMethod + '\script.jas : Function not valid: ' + cmd[1] + '(' + cmd[2] + ')');

        end
        else
        begin
          if (Cmd[0] = 'function') and (StrCount(Cmd[2], ',') = 0) then
          begin
            Params := Explode(',', Cmd[2]);

            if LowerCase(Cmd[1]) = 'toblackandwhite' then
            begin
              Captcha.ToBlackAndWhite(StrToFloat(Trim(Params[0])));
              Continue;
            end
            else
            if LowerCase(Cmd[1]) = 'reducewhitenoise' then
            begin
              Captcha.ReduceWhiteNoise(StrToInt(Trim(Params[0])));
              Continue;
            end
            else
            if LowerCase(Cmd[1]) = 'normalize' then
            begin
              Captcha.Normalize(StrToFloat(Trim(Params[0])));
              Continue;
            end;
            if LowerCase(Cmd[1]) = 'convertpixel' then
            begin
              Captcha.ConvertPixel(Trim(Params[0]));
              Continue;
            end
            else
            if LowerCase(Cmd[1]) = 'blurit' then
            begin
              Captcha.BlurIt(StrToInt(Trim(Params[0])));
              Continue;
            end
            else
              Logger.Warning('Command unimplemented (1): ' + Cmd[1]);
          end
          else
          if (Cmd[0] = 'function') and (StrCount(Cmd[2], ',') = 1) then
          begin
            Params := Explode(',', Cmd[2]);

            if LowerCase(Cmd[1]) = 'cleanbyrgbdistance' then
            begin
              Captcha.CleanByRGBDistance(StrToInt(Trim(Params[0])), StrToInt(Trim(Params[1])));
              Continue;
            end
            else
              Logger.Warning('Command unimplemented (2): ' + cmd[1]);
          end;
        end;

    end;
  except

  end;
end;

procedure TJACScript.ParseScriptFile;
var
  Script: string;
  Lines, PCmd: TArrayofstring;
  CaptchaPrepCmds, ScriptCmds, LetterCmds: TList<TArrayofstring>;
  StartAt, i: Integer;
begin
  if not FileExists(FMethod + FScriptFile) then
  begin
    Logger.Severe('script.jas not found!');
    Exit;
  end;

  Logger.Fine('Parsing Script.jas');

  Script := LoadStringFromFile(FMethod + FScriptFile);
  Script := StringReplace(Script, #13#10, #13, [rfReplaceAll]);
  Script := StringReplace(Script, #10#13, #13, [rfReplaceAll]);
  Script := StringReplace(Script, #10, #13, [rfReplaceAll]);
  Lines := Explode(#13, Script);

  CaptchaPrepCmds := TList<TArrayofstring>.Create;
  ScriptCmds := TList<TArrayofstring>.Create;
  LetterCmds := TList<TArrayofstring>.Create;

  for i := 0 to High(Lines) do
  begin
    Lines[i] := Trim(Lines[i]);

    // leere Zeile, oder Comment
    if (Pos('#', Lines[i]) = 1) or (Length(Lines[i]) < 4) then
      Continue;

    if Lines[i][Length(Lines[i])] <> ';' then
    begin
      Logger.Severe(FMethod + '\script.jas: Syntax error (; missing?) near line ' + IntToStr(i) + ': ' + Lines[i] + ' (last char: #' + IntToStr(Ord(Lines[i][Length(Lines[i])])) + ')');
      Exit;
    end;

    SetLength(Lines[i], Length(Lines[i]) - 1);

    StartAt := Pos('captcha.prepare.', Lines[i]);
    if StartAt > 0 then
    begin
      PCmd := ParseCommand(Copy(Lines[i], StartAt + 16, Length(Lines[i])));
      CaptchaPrepCmds.Add(PCmd);

      Continue;
    end;

    StartAt := Pos('param.', Lines[i]);
    if startAt > 0 then
    begin
      PCmd := ParseCommand(Copy(Lines[i], StartAt + 6, Length(Lines[i])));

      if PCmd[0] <> 'parameter' then
        Logger.Severe(FMethod + '\script.jas: Syntax (parameter1)  error near line ' + IntToStr(i) + ': ' + Lines[i]);

      ScriptCmds.Add(PCmd);

      Continue;
    end;

    StartAt := Pos('letter.prepare.', Lines[i]);
    if StartAt > 0 then
    begin
      PCmd := ParseCommand(Copy(Lines[i], StartAt + 15, Length(Lines[i])));
      LetterCmds.Add(PCmd);

      Continue;
    end;

    Logger.Severe(FMethod + '\script.jas: Syntax error near line ' + IntToStr(i) + ': ' + Lines[i]);
  end;

  FCaptchaPrepareCommands := CaptchaPrepCmds;
  FJacCommands := ScriptCmds;
  FLetterCommands := LetterCmds;
end;

function TJACScript.ParseCommand(Cmd: string): TArrayofstring;
var
  Tmp: TArrayofstring;
begin
  SetLength(Result, 3);
  Cmd := StringReplace(Cmd, ' ', '', [rfReplaceAll]);

  if Pos('=', Cmd) > 0 then
  begin
    Result[0] := 'parameter';
    Tmp := Explode('=', cmd);
    Result[1] := Tmp[0];
    Result[2] := Tmp[1];
    Tmp := nil;
  end
  else
  if (Pos('(', Cmd) > 0) and (Pos(')', Cmd) > 0) then
  begin
    Tmp := Explode('(', Cmd);
    Result[0] := 'function';
    Result[1] := Tmp[0];
    Result[2] := StringReplace(Tmp[1], ')', '', [rfReplaceAll]);
    Tmp := nil;
  end;
end;

procedure TJACScript.SetColorType(CType: string);
var
  Components: TCCArray;
  i: Integer;
begin
  SetParam('colorType', CType);
  for i := 0 to 2 do
    Components[i] := 3;
  FColorFaktor := 1;

  for i := Length(CType) downto 1 do
  begin
    if (CType[i] = 'h') then
    begin
      Components[i + 3 - Length(CType) - 1] := 0;
      FColorFaktor := FColorFaktor * 256;
      FColor := 0;
    end;

    if (CType[i] = 's') then
    begin
      Components[i + 3 - Length(CType) - 1] := 1;
      FColorFaktor := FColorFaktor * 256;
      FColor := 0;
    end;

    if (CType[i] = 'b') then
    begin
      Components[i + 3 - Length(CType) - 1] := 2;
      FColorFaktor := FColorFaktor * 256;
      FColor := 0;
    end;

    if (CType[i] = 'R') then
    begin
      Components[i + 3 - Length(CType) - 1] := 0;
      FColorFaktor := FColorFaktor * 256;
      FColor := 1;
    end;

    if (CType[i] = 'G') then
    begin
      Components[i + 3 - Length(CType) - 1] := 1;
      FColorFaktor := FColorFaktor * 256;
      FColor := 1;
    end;

    if (CType[i] = 'B') then
    begin
      Components[i + 3 - Length(CType) - 1] := 2;
      FColorFaktor := FColorFaktor * 256;
      FColor := 1;
    end;
  end;

  FColorComponents := Components;
end;

function TJACScript.ToType(Arg: string; Key: string): Variant;
var
  Current: Variant;
begin
  Current := GetParam(Key);

  if VarType(Current) in [varSmallInt, varInteger, varShortInt..varUInt64] then
    Exit(StrToInt(Arg));

  if (VarType(Current) = varString) or (VarType(Current) = varUString) then
    Exit(Arg);

  if VarType(Current) in [varSingle, varDouble, varCurrency] then
    Exit(StrToFloat(Arg));

  if VarType(Current) = varBoolean then
    Exit(LowerCase(Arg) = 'true');

  Logger.Severe('Parameter ' + Key + ' nicht initialisiert oder Typ unbekannt (' + IntToStr(Integer(VarType(Current))) + ')');
  Result := Null;
end;

end.
