(*
     Catcher! OCR Module  -  Port of JAntiCaptcha
     Copyright (C) 2010  CandiSoft

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

unit PixelObject;

interface

uses UTILITIES, Math, SysUtils, Generics.Collections, Generics.Defaults, untLogger;

type
  TPixelObject = class;
  TArrayofPixelObject = array of TPixelObject;

  TPixelObjectColor = record
    Color, Count: Integer;

    constructor Create(AColor: Integer);
  end;

  TPixelObject = class
  private
    FObject: TList<TArrayofInteger>;
    FAvg, FNoAvgChanges, FAvgIsSaveNum, FSaveAvg, FXMin, FXMax, FYMin, FYMax: Integer;
    FOwner: TObject;         // TPixelGrid
    FContrast, FWhiteContrast: Double;
    FLetterColor: Integer;
    FBordered: Boolean;
    FGrid: TDictionary<Integer, TDictionary<Integer, TArrayofInteger>>;
    FColors: TList<TPixelObjectColor>;

    function GetArea: Integer;
    function GetHeight: Integer;
    function GetSize: Integer;
    function GetWidth: Integer;
  public
    Detected: TObject;   // TLetterComperator

    constructor Create(AGrid: TObject);
    destructor Destroy; override;

    function AddColor(const Color: Integer): Boolean;
    procedure Add(X, Y, Color: Integer); overload;
    procedure Add(Current: TPixelObject); overload;
    function Align: TPixelObject; overload;
    function Align(AngleA, AngleB: Integer): TPixelObject; overload;
    function Cut(X1, X2, Overlap: Integer): TArrayofPixelObject;
    function DoesColorAverageFit(Color: Integer): Boolean;
    function ElementAt(i: Integer): TArrayofInteger;
    function GetBorderVector(Letter: TObject): T2DArrayofInteger;
    //function getDistanceTo()
    function GetLocation: TArrayofInteger;
    function GetMassValue(X, Y: Integer): Int64;
    function GetWidthToHeight: Double;
    function Split(SplitNum: Integer): TList<TPixelObject>; overload;
    function Split(SplitNum, Overlap: Integer): TList<TPixelObject>; overload;
    function SplitAt(Position: Integer): TArrayofPixelObject;
    function ToLetter: TObject;
    function ToString: string; override;
    function Turn(Angle: Double): TPixelObject;

    function GetMostColor: Integer;
    function IsTouching(B: TPixelObject; FollowX: Boolean; RadiusX, RadiusY: Integer): Boolean; overload;
    function IsTouching(X, Y: Integer; FollowX: Boolean; RadiusX, RadiusY: Integer): Boolean; overload;

    property Area: Integer read GetArea;
    property Average: Integer read FAvg;
    property Bordered: Boolean read FBordered write FBordered;
    property Colors: TList<TPixelObjectColor> read FColors;
    property Contrast: Double read FContrast write FContrast;
    property Grid: TDictionary<Integer, TDictionary<Integer, TArrayofInteger>> read FGrid;
    property Height: Integer read GetHeight;
    property Obj: TList<TArrayofInteger> read FObject;
    property Owner: TObject read FOwner;
    property Size: Integer read GetSize;
    property WhiteContrast: Double read FWhiteContrast write FWhiteContrast;
    property Width: Integer read GetWidth;
    property XMin: Integer read FXMin;
    property YMin: Integer read FYMin;
    property XMax: Integer read FXMax;
    property YMax: Integer read FYMax;
  end;

function PixelObjectComparer(const Left, Right: TPixelObject): Integer;

implementation

uses PixelGrid, AntiCaptcha, Letter, LetterComperator;

{ TPixelObjectColor }

constructor TPixelObjectColor.Create(AColor: Integer);
begin
  Color := AColor;
  Count := 1;
end;

function PixelObjectColorComparer(const Left, Right: TPixelObjectColor): Integer;
begin
  if Left.Count < Right.Count then
    Result := -1
  else
  if Left.Count > Right.Count then
    Result := 1
  else
    Result := 0;
end;

function PixelObjectComparer(const Left, Right: TPixelObject): Integer;
begin
  if Left.GetLocation[0] < Right.GetLocation[0] then
    Result := -1
  else
  if Left.GetLocation[0] > Right.GetLocation[0] then
    Result := 1
  else
    Result := 0;
end;

{ TPixelObject }

constructor TPixelObject.Create(AGrid: TObject);
begin
  FOwner := AGrid;
  FGrid := TDictionary<Integer, TDictionary<Integer, TArrayofInteger>>.Create;
  FColors := TList<TPixelObjectColor>.Create(
    TComparer<TPixelObjectColor>.Construct(PixelObjectColorComparer));
  FObject := TList<TArrayofInteger>.Create;

  FAvg := 0;
  FAvgIsSaveNum := 10;
  FNoAvgChanges := 0;
  FSaveAvg := 0;
  FXMin := MAXINT;
  FXMax := MININT;
  FYMin := MAXINT;
  FYMax := MININT;
  FWhiteContrast := 1;
  FLetterColor := 0;
  FBordered := True;
end;

destructor TPixelObject.Destroy;
var
  Y: TDictionary<Integer, TArrayofInteger>;
  i: Integer;
begin
  for Y in FGrid.Values do
  begin
    for i in Y.Keys do
      Y[i] := nil;

    Y.Free;
  end;
  FGrid.Free;
  FColors.Free;
  for i := 0 to FObject.Count - 1 do
    FObject[i] := nil;
  FObject.Free;

  inherited;
end;

function TPixelObject.AddColor(const Color: Integer): Boolean;
var
  POC: TPixelObjectColor;
  i: Integer;
begin
  POC := TPixelObjectColor.Create(Color);

  for i := 0 to FColors.Count - 1 do
    if TPixelObjectColor(FColors[i]).Color = Color then
    begin
      POC := FColors[i];
      Inc(POC.Count);
      FColors[i] := POC;
      Exit(False);
    end;

  FColors.Add(POC);
  Result := True;
end;

function TPixelObject.GetMostColor: Integer;
begin
  FColors.Sort;
  if FColors.Count > 0 then
    Result := TPixelObjectColor(FColors[0]).Color
  else
    Result := FAvg;
end;

procedure TPixelObject.Add(X: Integer; Y: Integer; Color: Integer);
var
  Tmp: TArrayofInteger;
  TmpAvg: Integer;
  Row: TDictionary<Integer, TArrayofInteger>;
begin
  SetLength(Tmp, 3);
  Tmp[0] := X;
  Tmp[1] := Y;
  Tmp[2] := Color;
  TmpAvg := FAvg;

  if Color >= 0 then
  begin
    FAvg := MixColors(FAvg, Color, getSize, 1);
    AddColor(Color);
  end;

  if FGrid.ContainsKey(X) then
    Row := FGrid[X]
  else
    Row := nil;

  if Row = nil then
  begin
    Row := TDictionary<Integer, TArrayofInteger>.Create;
    FGrid.AddOrSetValue(X, Row);
  end;

  Row.AddOrSetValue(Y, Copy(Tmp, 0, 3));

  if Abs(FAvg - TmpAvg) < (FOwner as TPixelGrid).GetMaxPixelValue * FContrast then
  begin
     Inc(FNoAvgChanges);

     if (FAvgIsSaveNum <= FNoAvgChanges) and (FSaveAvg = 0) then
       FSaveAvg := FAvg
       // Logger.Info('SaveAvg: ' + IntToStr(FAvg));

  end
  else
    FNoAvgChanges := 0;

  FObject.Add(Tmp);

  FXMin := Math.Min(X, FXMin);
  FXMax := Math.Max(X, FXMax);
  FYMin := Math.Min(Y, FYMin);
  FYMax := Math.Max(Y, FYMax);
end;

procedure TPixelObject.Add(Current: TPixelObject);
var
  i: Integer;
  Tmp: TArrayofInteger;
begin
  FAvg := MixColors(FAvg, Current.Average, getSize, Current.getSize);
  FXMin := Math.Min(Current.XMin, FXMin);
  FXMax := Math.Max(Current.XMax, FXMax);
  FYMin := Math.Min(Current.YMin, FYMin);
  FYMax := Math.Max(Current.YMax, FYMax);

  for i in Current.Grid.Keys do
    FGrid.AddOrSetValue(i, Current.Grid[i]);

  for i := 0 to Current.Obj.Count - 1 do
  begin
    SetLength(Tmp, 3);
    Tmp[0] := Current.Obj[i][0];
    Tmp[1] := Current.Obj[i][1];
    Tmp[2] := -1;
    FObject.Add(Tmp);
  end;
end;

function TPixelObject.Align: TPixelObject;
var
  Accuracy, Angle, Steps: Integer;
  R, L, Next: TPixelObject;
begin
  Accuracy := 1;
  R := Turn(-Accuracy);
  L := Turn(Accuracy);

  if (R.GetWidthToHeight >= GetWidthToHeight) and (L.GetWidthToHeight >= GetWidthToHeight) then
    Exit(Self);

  if R.GetWidthToHeight < L.GetWidthToHeight then
    Steps := -Accuracy
  else
    Steps := Accuracy;

  Angle := Steps * 2;

  if R.GetWidthToHeight < L.GetWidthToHeight then
    Result := R
  else
    Result := L;

  Next := Turn(Angle);
  while Next.GetWidthToHeight < Result.GetWidthToHeight do
  begin
    Result := Next;

    Angle := Angle + Steps;

    Next := Turn(Angle);
  end;
end;

function TPixelObject.Align(AngleA: Integer; AngleB: Integer): TPixelObject;
var
  Tmp, Accuracy, Angle: Integer;
  BestValue: Double;
  PTmp: TPixelObject;
begin
  if AngleB < AngleA then
  begin
    Tmp := AngleB;
    AngleB := AngleA;
    AngleA := Tmp;
  end;

  Accuracy := TAntiCaptcha((FOwner as TPixelGrid).Owner).Script.GetParam('AlignAngleSteps');
  BestValue := MaxDouble;

  Result := nil;

  Angle := AngleA;
  while Angle < AngleB do
  begin
    if Angle < 0 then
      PTmp := Turn(360 + Angle)
    else
      PTmp := Turn(Angle);

    if PTmp.getWidthToHeight < BestValue then
    begin
      BestValue := PTmp.getWidthToHeight;
      Result := PTmp;
    end;

    Angle := Angle + Accuracy;
  end;
end;

function TPixelObject.Cut(X1: Integer; X2: Integer; Overlap: Integer): TArrayofPixelObject;
var
  Pre, Post, Cutter: TPixelObject;
  i: Integer;
  Akt: TArrayofInteger;
begin
  Pre := TPixelObject.Create(FOwner);
  Post := TPixelObject.Create(FOwner);
  Cutter := TPixelObject.Create(FOwner);

  if (X1 < Integer(TAntiCaptcha(((FOwner) as TPixelGrid).Owner).Script.GetParam('minimumLetterWidth'))) then
    FreeAndNil(Pre);

  if (FXMax - FXMin - X2 < Integer(TAntiCaptcha(((FOwner) as TPixelGrid).Owner).Script.GetParam('minimumLetterWidth'))) then
    FreeAndNil(Post);

  for i := 0 to getSize - 1 do
  begin
    Akt := ElementAt(i);
    if (Akt[0] >= FXMin + X1 - Overlap) and (Akt[0] <= (FXMin + X2) + Overlap) then
      Cutter.Add(Akt[0], Akt[1], FSaveAvg);

    if (Pre <> nil) and (Akt[0] < FXMin + X1) then
      Pre.Add(Akt[0], Akt[1], FSaveAvg)
    else
      if (Pre = nil) and (Akt[0] < FXMin + X1) then
        Cutter.Add(Akt[0], Akt[1], FSaveAvg);

    if (Post <> nil) and (Akt[0] > (FXMin + X2)) then
      Post.Add(Akt[0], Akt[1], FSaveAvg)
    else
      if (Post = nil) and (Akt[0] > (FXMin + X2)) then
        Cutter.Add(Akt[0], Akt[1], FSaveAvg);
  end;

  SetLength(Result, 3);
  Result[0] := Pre;
  Result[1] := Cutter;
  Result[2] := Post;
end;

function TPixelObject.DoesColorAverageFit(Color: Integer): Boolean;
var
  TAvg: Integer;
begin
  if GetSize > 50000 then
  begin
    Logger.Severe('Objekt scheint sehr gro zu werden. objectColorContrast zu hoch?');
    Exit(False);
  end;

  if FSaveAvg = 0 then
    TAvg := FAvg
  else
    TAvg := FSaveAvg;

  // Filtere zu helle Pixel aus
  if Color > Trunc(FWhiteContrast * (FOwner as TPixelGrid).getMaxPixelValue) then
    Exit(False);

  if GetSize = 0 then
    Exit(True);

  Result := Abs(TAvg - Color) < (FOwner as TPixelGrid).getMaxPixelValue * FContrast;
end;

function TPixelObject.ElementAt(i: Integer): TArrayofInteger;
begin
  Result := FObject[i];
end;

function TPixelObject.GetArea: Integer;
begin
  Result := GetWidth * GetHeight;
end;

function TPixelObject.GetBorderVector(Letter: TObject): T2DArrayofInteger;
var
  AGrid, Map: T2DArrayofInteger;
  i, X, Y, XX, YY: Integer;
  Akt: TArrayofInteger;
  C: Boolean;
begin
  if Letter = nil then
    Letter := Self.ToLetter;

  Result := nil;

  AGrid := TLetter(Letter).Grid;
  for i := 0 to GetSize - 1 do
  begin
    Akt := ElementAt(i);
    X := Akt[0];
    Y := Akt[1];

    Map := TLetter(Letter).GetLocalMap(AGrid, X, Y);
    C := False;
    for XX := 0 to 2 do
      for YY := 0 to 2 do
        if Map[XX][YY] <> 0 then
        begin
          C := True;
          Break;
        end;

    if C then
    begin
      SetLength(Result, Length(Result) + 1);
      Result[High(Result)] := akt;
    end;
  end;
end;

function TPixelObject.GetHeight: Integer;
begin
  Result := FYMax - FYMin + 1;
end;

function TPixelObject.GetLocation: TArrayofInteger;
begin
  SetLength(Result, 2);
  Result[0] := FXMin;
  Result[1] := FYMin;
end;

function TPixelObject.GetMassValue(X: Integer; Y: Integer): Int64;
var
  i, Col: Integer;
  Akt: TArrayofInteger;
  Dist: Double;
begin
  Result := 0;

  for i := 0 to GetSize - 1 do
  begin
    Akt := ElementAt(i);
    if (X = Akt[0]) and (Y = Akt[1]) then
      Result := Result + (FOwner as TPixelGrid).getMaxPixelValue * 2
    else
    begin
      Col := (FOwner as TPixelGrid).getMaxPixelValue - Akt[2];
      Dist := Sqrt(Abs(Akt[0] - X) * Abs(Akt[0] - X) + Abs(Akt[1] - Y) * Abs(Akt[1] - Y)) / 0.5;

      Result := Trunc(Col / (Dist + 1));
    end;
  end;
end;

function TPixelObject.GetSize: Integer;
begin
  Result := FObject.Count;
end;

function TPixelObject.GetWidth: Integer;
begin
  Result := FXMax - FXMin + 1;
end;

function TPixelObject.GetWidthToHeight: Double;
begin
  Result := GetWidth / GetHeight;
end;

function TPixelObject.IsTouching(B: TPixelObject; FollowX: Boolean; RadiusX,
  RadiusY: Integer): Boolean;
var
  Row: TDictionary<Integer, TArrayofInteger>;
  PX: TArrayofInteger;
  SX, SY: Integer;
begin
  if (B.XMin > FXMax + RadiusX) or
     (B.XMax < FXMin - RadiusY) or
     (B.YMin > FYMax + RadiusX) or
     (B.YMax < FYMin - RadiusY) then
    Exit(False);

  for PX in B.Obj do
    for SX := -RadiusX to RadiusX do
    begin
      if FGrid.ContainsKey(PX[0] + SX) then
        Row := FGrid[PX[0] + SX]
      else
        Row := nil;

      if Row = nil then
        Continue;

      for SY := -RadiusY to RadiusY do
      begin
        if (Abs(SX) = Abs(SY)) and (not FollowX) then
          Continue;

        if (Row.ContainsKey(PX[1] + SY)) and (Row[PX[1] + SY] <> nil) then
          Exit(True);
      end;
    end;

  Result := False;
end;

function TPixelObject.IsTouching(X, Y: Integer; FollowX: Boolean; RadiusX,
  RadiusY: Integer): Boolean;
var
  Row: TDictionary<Integer, TArrayofInteger>;
  SX, SY: Integer;
begin
  if (X > FXMax + RadiusX) or
     (X < FXMin - RadiusY) or
     (Y > FYMax + RadiusX) or
     (Y < FYMin - RadiusY) then
    Exit(False);

  for SX := -RadiusX to RadiusX do
  begin
    if FGrid.ContainsKey(X + SX) then
      Row := FGrid[X + SX]
    else
      Row := nil;

    if Row = nil then
      Continue;

    for SY := -RadiusY to RadiusY do
    begin
      if (Abs(SX) = Abs(SY)) and (not FollowX) then
        Continue;

      if (Row.ContainsKey(Y + SY)) and (Row[Y + SY] <> nil) then
        Exit(True);
    end;
  end;

  Result := False;
end;

function TPixelObject.Split(SplitNum: Integer): TList<TPixelObject>;
begin
  Result := Split(SplitNum, 0);
end;

function TPixelObject.Split(SplitNum: Integer; Overlap: Integer): TList<TPixelObject>;
var
  t, Part, i, x: Integer;
  Akt: TArrayofInteger;
begin
  Result := TList<TPixelObject>.Create;

  for t := 0 to SplitNum - 1 do
    Result.Add(TPixelObject.Create(FOwner));

  Part := GetWidth div SplitNum;

  for i := 0 to GetSize - 1 do
  begin
    Akt := ElementAt(i);

    for x := 0 to SplitNum - 1 do
      if (Akt[0] >= FXMin + x * Part - Overlap) and (Akt[0] <= (FXMin + (x + 1) * Part) + Overlap) then
        TPixelObject(Result[x]).Add(Akt[0], Akt[1], FSaveAvg);
  end;
end;

function TPixelObject.SplitAt(Position: Integer): TArrayofPixelObject;
var
  i, x: Integer;
  Akt: TArrayofInteger;
  b: Boolean;
begin
  SetLength(Result, 2);
  Result[0] := TPixelObject.Create(FOwner);
  Result[1] := TPixelObject.Create(FOwner);

  for i := 0 to GetSize - 1 do
  begin
    Akt := ElementAt(i);
    b := True;

    for x := 0 to 1 do
    begin
      if (Akt[0] >= FXMin + x * Position) and (Akt[0] <= (FXMin + (x + 1) * Position)) then
      begin
        Result[x].Add(Akt[0], Akt[1], FSaveAvg);
        b := False;
      end;
    end;

    if b then
      Result[1].Add(Akt[0], Akt[1], FSaveAvg);
  end;
end;

function TPixelObject.ToLetter: TObject;
var
  Res: T2DArrayofInteger;
  i, X, Y: Integer;
  Akt: TArrayofInteger;
  L: TLetter;
  Loc: TStatic01ArrayofInt;
begin
  SetLength(Res, GetWidth, GetHeight);

  for X := 0 to GetWidth - 1 do
    for Y := 0 to GetHeight - 1 do
      Res[X][Y] := (FOwner as TPixelGrid).getMaxPixelValue;

  for i := 0 to GetSize - 1 do
  begin
    Akt := ElementAt(i);
    Res[Akt[0] - FXMin][Akt[1] - FYMin] := FAvg;
  end;

  L := TLetter((FOwner as TPixelGrid).CreateLetter);
  L.SetElementPixel(getSize);

  Loc[0] := FXMin;
  Loc[1] := FYMin;
  L.Location := Loc;

  L.Grid := Res;
  L.Detected := Detected;
  Result := L;
end;

function TPixelObject.ToString: string;
begin
  Result := 'Debug: ' + IntToStr(Self.GetLocation[0]);
end;

function TPixelObject.Turn(Angle: Double): TPixelObject;
var
  i: Integer;
  Akt, N: TArrayofInteger;
begin
  Result := TPixelObject.Create(FOwner);

  for i := 0 to GetSize - 1 do
  begin
    Akt := ElementAt(i);
    N := TurnCoordinates(Akt[0], Akt[1], FXMin + GetWidth div 2, FYMin + GetHeight div 2, Angle);
    Result.Add(N[0], N[1], FAvg);
  end;
end;

end.
