(*
     Catcher! OCR Module  -  Port of JAntiCaptcha
     Copyright (C) 2010  CandiSoft

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

unit Captcha;

interface

uses Windows, SysUtils, PixelGrid, Letter, UTILITIES, Graphics, LetterComperator,
     Math, untLogger, PixelObject, Generics.Collections;

type
  TArrayofLetter = array of TLetter;
  TArrayofBoolean = array of Boolean;
  TArrayofLetterComperator = array of TLetterComperator;

  TCaptcha = class(TPixelGrid)
  private
    FLastLetterX: Integer;
    FPrepared: Boolean;
    FSeparatedLetters: TArrayofLetter;
    FGaps: TArrayofBoolean;
    FValityPercent: Double;
    FPerfectObjectDetection: Boolean;
    FRGBGrid: T2DArrayofInteger;

    function GetNextLetter(letterID: Integer): TLetter;
    procedure MergeMultiPartObjects(var Objects: TList<TPixelObject>);
  public
    class function GetCaptcha(FileName: string; AOwner: TObject): TCaptcha; overload;
    class function GetCaptcha(Image: TBitmap; AOwner: TObject): TCaptcha; overload;
    class function GetCaptcha(A, B: TLetter): TCaptcha; overload;

    constructor Create(AWidth, AHeight: Integer);
    destructor Destroy; override;

    procedure CleanBackgroundByHorizontalSampleLine(X1, X2, Y1, Y2: Integer);
    procedure CleanWithDetailMask(Mask: TCaptcha; Dif: Integer);
    procedure CleanWithMask(Mask: TCaptcha; AWidth, AHeight: Integer);
    procedure Concat(Tmp: TCaptcha);
    procedure ConvertOriginalPixel(NewColorFormat: string);
    procedure ConvertPixel(NewColorFormat: string);
    function GetAverage(PX, PY, AWidth, AHeight: Integer; Mask: TCaptcha): Integer; overload;
    function GetBiggestObjects(LetterNum, MinArea: Integer; Contrast, ObjectContrast: Double): TList<TPixelObject>;

    procedure SetPixel(APixel: TArrayofInteger);

    function GetSimplified(Faktor: Double): TCaptcha;
    function GetLetters(LetterNum: Integer; Contrast, ObjectContrast: Double; MinArea: Integer): TArrayofLetter; overload;
    function Reset: Boolean;
    function GetLetters(LetterNum: Integer): TArrayofLetter; overload;
    function GetLetters(LetterNum: Integer; AGaps: TArrayofInteger): TArrayofLetter; overload;
    function GetLetters0(LetterNum: Integer): TArrayofLetter; overload;

    procedure CleanByRGBDistance(Mode, Tollerance: Integer);

    property IsPrepared: Boolean read FPrepared write FPrepared;
    property PerfectObjectDetection: Boolean read FPerfectObjectDetection write FPerfectObjectDetection;
    property ValityPercent: Double read FValityPercent write FValityPercent;
  end;

implementation

uses AntiCaptcha {$IFDEF DEBUG}, GUI{$ENDIF};

class function TCaptcha.GetCaptcha(FileName: string; AOwner: TObject): TCaptcha;
var
  Img: TBitmap;
begin
  Img := TBitmap.Create;
  Img.LoadFromFile(FileName);
  Result := GetCaptcha(Img, AOwner);
end;

class function TCaptcha.GetCaptcha(Image: TBitmap; AOwner: TObject): TCaptcha;
var
  X, Y: Integer;
begin
  Result := nil;
  if (Image.Width <= 0) or (Image.Height <= 0) then
  begin
    Logger.Severe('ERROR: Image nicht korrekt. Kein Inhalt. Pfad Angaben korrigieren!');
    Exit;
  end;

  Result := TCaptcha.Create(Image.Width, Image.Height);
  Result.Owner := AOwner;
  Logger.Fine(IntToStr(Image.Width) + ' | ' + IntToStr(Image.Height));

  for X := 0 to Image.Width - 1 do
    for Y := 0 to Image.Height - 1 do
      Result.SetPixelValue(X, Y, Image.Canvas.Pixels[X, Y]);

  Result.SetOrgGrid(TPixelGrid.GetGridCopy(Result.Grid));
end;

class function TCaptcha.GetCaptcha(A: TLetter; B: TLetter): TCaptcha;
var
  NewWidth, NewHeight: Integer;
  X, Y: Integer;
begin
  NewWidth := A.Width + B.Width + 6;
  NewHeight := Max(A.Height, B.Height);

  Result := TCaptcha.Create(NewWidth, NewHeight);

  if A.Owner <> nil then
    Result.Owner := A.Owner;
  if Result.Owner = nil then
    Result.Owner := B.Owner;
  if Result.Owner = nil then
    logger.warning('Owner konnte nicht bestimmt werden! Dieser Captcha ist nur eingeschrnkt verwendbar.');

  for X := 0 to A.Width - 1 do
    for Y := 0 to NewHeight - 1 do
      if Y < A.Height then
        Result.Grid[X][Y] := A.Grid[X][Y]
      else
        Result.Grid[X][Y] := A.GetMaxPixelValue;

  for X := A.Width to A.Width + 5 do    // + 6 - 1
    for Y := 0 to NewHeight - 1 do
      if (X = A.Width + 2) or (X = A.Width + 3) then
        Result.Grid[X][Y] := 0
      else
        Result.Grid[X][Y] := A.GetMaxPixelValue;

  for X := A.Width + 6 to NewWidth - 1 do
    for Y := 0 to NewHeight - 1 do
      if Y < B.Height then
        Result.Grid[X][Y] := B.Grid[X - (A.Width + 6)][Y]
      else
        Result.Grid[X][Y] := B.GetMaxPixelValue;
end;

constructor TCaptcha.Create(AWidth, AHeight: Integer);
begin
  SetLength(FGrid, AWidth, AHeight);
  SetLength(FRGBGrid, AWidth, AHeight);
end;

destructor TCaptcha.Destroy;
begin
  // xxx

  inherited;
end;

procedure TCaptcha.CleanBackgroundByHorizontalSampleLine(X1, X2, Y1, Y2: Integer);
begin
  CleanBackgroundByColor(GetAverage(X1, Y1, X2 - X1, Y2 - Y1));
end;

procedure TCaptcha.CleanWithDetailMask(Mask: TCaptcha; Dif: Integer);
var
  NewGrid, Test: T2DArrayofInteger;
  X, Y: Integer;
begin
  SetLength(NewGrid, GetWidth, GetHeight);
  SetLength(Test, GetWidth, GetHeight);

  if (Mask.GetWidth <> GetWidth) or (Mask.Height <> GetHeight) then
  begin
    Logger.Info('ERROR Maske und Bild passen nicht zusammmen');
    Exit;
  end;

  Logger.Info('Dif: ' + IntToStr(Dif));

  for X := 0 to GetWidth - 1 do
    for Y := 0 to GetHeight - 1 do
    begin
      Test[X][Y] := Abs(Mask.GetPixelValue(X, Y) - GetPixelValue(X, Y));
      if GetColorDifference(Mask.GetPixelValue(X, Y), getPixelValue(X, Y)) < Dif then
        SetPixelValue(X, Y, NewGrid, GetMaxPixelValue)
      else
        Newgrid[X][Y] := FGrid[X][Y];
    end;

  FGrid := NewGrid;
end;

procedure TCaptcha.CleanWithMask(Mask: TCaptcha; AWidth: Integer; AHeight: Integer);
var
  NewGrid: T2DArrayofInteger;
  X, Y: Integer;
begin
  SetLength(NewGrid, GetWidth, GetHeight);

  if (Mask.Width <> GetWidth) or (Mask.Height <> GetHeight) then
  begin
    Logger.Info('ERROR Maske und Bild passen nicht zusammmen');
    Exit;
  end;

  for X := 0 to GetWidth - 1 do
    for Y := 0 to GetHeight - 1 do
    begin
      if (Mask.GetPixelValue(X, Y) < 100) then
        SetPixelValue(X, Y, NewGrid, GetAverage(X, Y, AWidth, AHeight, Mask))
      else
        NewGrid[X][Y] := FGrid[x][y];
    end;

  FGrid := NewGrid;
end;

procedure TCaptcha.Concat(Tmp: TCaptcha);
var
  X, Y: Integer;
begin
  if (GetWidth <> Tmp.Width) or (GetHeight <> Tmp.Height) then
  begin
    Logger.Severe('Concat fehlgeschlagen: Dimension nicht gleich');
    Exit;
  end;

  for X := 0 to GetWidth - 1 do
    for Y := 0 to GetHeight - 1 do
      if (GetPixelValue(X, Y, Tmp.Grid) > Double(TAntiCaptcha(FOwner).Script.GetParam('getBlackPercent')) * GetMaxPixelValue) then
        SetPixelValue(X, Y, MixColors(GetPixelValue(X, Y), GetPixelValue(X, Y, Tmp.Grid)));
end;

procedure TCaptcha.ConvertOriginalPixel(NewColorFormat: string);
var
  Y, X: Integer;
  RGB: TArrayofInteger;
  Col: TArrayofDouble;
begin
  TAntiCaptcha(FOwner).Script.SetColorType(NewColorFormat);
  for Y := 0 to GetHeight - 1 do
    for X := 0 to GetWidth - 1 do
    begin
      SetLength(Col, 4);

      if (FOwner as TAntiCaptcha).Script.ColorFormat = 0 then
      begin
        SetLength(RGB, 3);
        RGB[0] := GetRValue(FRGBGrid[X][Y]);
        RGB[1] := GetGValue(FRGBGrid[X][Y]);
        RGB[2] := GetBValue(FRGBGrid[X][Y]);
        Col := RGBToHSB(rgb);
      end
      else
        if (FOwner as TAntiCaptcha).Script.ColorFormat = 1 then
        begin
          Col[0] := GetRValue(FRGBGrid[X][Y]) / 255;
          Col[1] := GetGValue(FRGBGrid[X][Y]) / 255;
          Col[2] := GetBValue(FRGBGrid[X][Y]) / 255;
          Col[3] := 0;
        end;

      FGrid[X][Y] := Trunc(
                      (Col[(FOwner as TAntiCaptcha).Script.ColorComponents[0]] * 255) * 65536 +
                      (Col[(FOwner as TAntiCaptcha).Script.ColorComponents[1]] * 255) * 256   +
                      (Col[(FOwner as TAntiCaptcha).Script.ColorComponents[2]] * 255)
                    );
    end;
end;

procedure TCaptcha.ConvertPixel(NewColorFormat: string);
var
  Y, X: Integer;
  RGB: TArrayofInteger;
  Col: TArrayofDouble;
begin
  (FOwner as TAntiCaptcha).Script.SetColorType(NewColorFormat);
  for Y := 0 to GetHeight - 1 do
    for X := 0 to GetWidth - 1 do
    begin
      SetLength(Col, 4);

      if (FOwner as TAntiCaptcha).Script.ColorFormat = 0 then
      begin
        SetLength(RGB, 3);
        RGB[0] := GetRValue(FGrid[x][y]);
        RGB[1] := GetGValue(FGrid[x][y]);
        RGB[2] := GetBValue(FGrid[x][y]);
        Col := RGBToHSB(RGB);
      end
      else
        if (FOwner as TAntiCaptcha).Script.ColorFormat = 1 then
        begin
          Col[0] := GetRValue(FGrid[x][y]) / 255;
          Col[1] := GetGValue(FGrid[x][y]) / 255;
          Col[2] := GetBValue(FGrid[x][y]) / 255;
          Col[3] := 0;
        end;

      FGrid[x][y] := Trunc(
                      (Col[(FOwner as TAntiCaptcha).Script.ColorComponents[0]] * 255) * 65536 +
                      (Col[(FOwner as TAntiCaptcha).Script.ColorComponents[1]] * 255) * 256   +
                      (Col[(FOwner as TAntiCaptcha).Script.ColorComponents[2]] * 255)
                    );
    end;
end;

function TCaptcha.GetAverage(PX, PY, AWidth, AHeight: Integer; Mask: TCaptcha): Integer;
var
  avg, bv: TArrayofInteger;
  i, x, y, halfW, halfH: Integer;
begin
  i := 0;
  SetLength(avg, 3);
  avg[0] := 0;
  avg[1] := 0;
  avg[2] := 0;
  halfW := AWidth div 2;
  halfH := AHeight div 2;

  if (AWidth = 1) and (px = 0) then
    AWidth := 2;
  if (AHeight = 1) and (py = 0) then
    AHeight := 2;

  for x := Math.Max(0, px - halfW) to Math.Min(px + AWidth - halfW, GetWidth) - 1 do
    for y := Math.Max(0, py - halfH) to Math.Min(py + AHeight - halfH, GetHeight) - 1 do
    begin
      if Mask.GetPixelValue(x, y) > 100 then
      begin
        bv := hexToRGB(GetPixelValue(x, y));
        avg[0] := avg[0] + bv[0];
        avg[1] := avg[1] + bv[1];
        avg[2] := avg[2] + bv[2];

        Inc(i);
        bv := nil;
      end;
    end;

  avg[0] := avg[0] div i;
  avg[1] := avg[1] div i;
  avg[2] := avg[2] div i;
  Result := RGB(avg[0], avg[1], avg[2]);
end;

function TCaptcha.GetBiggestObjects(LetterNum: Integer; MinArea: Integer; Contrast: Double; ObjectContrast: Double): TList<TPixelObject>;
var
  Splitter, SplitNum, Found, i, MinWidth, MaxWidth: Integer;
  Objects, ObjectsRes, SplitObjects: TList<TPixelObject>;
  BX, BY, ObjectI, b, r, t, s, j: Integer;
  PixelObject, Current, Cutter, Pre, Post, PO, Next: TPixelObject;
  Letter: TLetter;
  ResLetter: TLetterComperator;
  SPObjects: TArrayofPixelObject;
  PerfectObjDetection: Boolean;
begin
  Found := 0;
  i := 0;
  // Alle Objekte aus dem captcha holen. Sie sind nach der Gre Sortiert
  Objects := GetObjects(Contrast, ObjectContrast);

  if Boolean(TAntiCaptcha(FOwner).Script.GetParam('directLetterDetection')) then
  begin
    logger.info('Get directLetterDetection');

    BX := TAntiCaptcha(owner).Script.GetParam('borderVarianceX');
    BY := TAntiCaptcha(owner).Script.GetParam('borderVarianceY');

    for ObjectI := 0 to Objects.Count - 1 do
    begin
      PixelObject := Objects[i];

      if PixelObject.Area > MinArea * 2 then
      begin
        TAntiCaptcha(FOwner).Script.SetParam('borderVarianceX', BX * TAntiCaptcha(owner).Script.getParam('minimumLetterWidth') * 100);
        TAntiCaptcha(FOwner).Script.SetParam('borderVarianceY', BY * TAntiCaptcha(owner).Script.getParam('minimumLetterWidth') * 100);
        Current := PixelObject;
        Letter := TLetter(Current.ToLetter);
        ResLetter := TAntiCaptcha(owner).GetLetter(Letter);
        ResLetter.ValityPercent := ResLetter.ValityPercent * 0.75;
        Current.detected := ResLetter;

        if (ResLetter.B <> nil) and (ResLetter.getDecodedValue <> '-') then
        begin
          b := PixelObject.Area - ResLetter.B.Area;
          if (b > MinArea) and (b > ResLetter.B.Area / 3) and
             (ResLetter.Offset <> nil) and (Length(ResLetter.Offset) > 0) then
          begin
            logger.info('dld: got letter: ' + ResLetter.GetDecodedValue);

            SPObjects := Current.Cut(ResLetter.Offset[0], ResLetter.Offset[0] + ResLetter.B.Width, Integer(TAntiCaptcha(owner).Script.GetParam('splitPixelObjectsOverlap')));
            Cutter := SPObjects[1];
            Pre := SPObjects[0];
            Post := SPObjects[2];

            {$IFDEF DEBUG}
            frmGUI.ShowImage(TLetter(Cutter.ToLetter).GetImage, 'cutter' + ResLetter.GetDecodedValue);
            {$ENDIF}

            ObjectsRes.Add(Cutter);
            Cutter.Detected := ResLetter;

            if (Pre <> nil) and (Pre.Area > MinArea) and (Pre.Size > TAntiCaptcha(FOwner).Script.GetParam('minObjectSize')) then
            begin
              Objects.Add(Pre);
              {$IFDEF DEBUG}
              frmGUI.ShowImage(TLetter(pre.toLetter).getImage, 'pre cut ' + resletter.getDecodedValue);
              {$ENDIF}
            end;

            if (Post <> nil) and (Post.Area > MinArea) and (Post.Size > TAntiCaptcha(FOwner).Script.GetParam('minObjectSize')) then
            begin
              Objects.Add(Post);
              {$IFDEF DEBUG}
              frmGUI.ShowImage(TLetter(post.ToLetter).GetImage, 'post cut ' + resletter.getDecodedValue);
              {$ENDIF}
            end;
          end;

        end
        else
        begin
          ObjectsRes.Add(Current);
          {$IFDEF DEBUG}
          frmGUI.ShowImage(TLetter(current.toLetter).getImage, 'normal ext ' + resletter.getDecodedValue);
          {$ENDIF}
        end;

        TAntiCaptcha(FOwner).Script.SetParam('borderVarianceX', BX);
        TAntiCaptcha(FOwner).Script.SetParam('borderVarianceY', BY);
      end
      else
      begin
        Current := PixelObject;
        Letter := TLetter(Current.ToLetter);
        ResLetter := TAntiCaptcha(FOwner).GetLetter(Letter);
        Current.Detected := ResLetter;

        // xxx
      end;
    end;
  end;

  MergeMultiPartObjects(Objects);

  if Boolean(TAntiCaptcha(FOwner).Script.GetParam('autoLetterNum')) then
  begin
    r := 0;

    if Objects.Count > 0 then
      for i := 0 to Objects.Count - 1 do
        if (TPixelObject(Objects[i]).Area > minArea) then
          Inc(r);

    TAntiCaptcha(owner).LetterNum := r;
    LetterNum := r;
  end;

  PerfectObjDetection := True;

  i := 0;
  while (i < Objects.Count) and (TPixelObject(Objects[i]).Area > MinArea) and (Found < LetterNum) do
  begin
    Inc(i);
    Logger.Info(IntToStr(TPixelObject(Objects[i - 1]).Width) + ' Element: ' + IntToStr(Found) + ' : ' + IntToStr(TPixelObject(Objects[i - 1]).Area));
    Inc(Found);
  end;

  if not TAntiCaptcha(owner).Script.GetParam('autoLetterNum') then
  begin
    Logger.Fine(Format('Found %d MinArea: %d', [Found, MinArea]));

    while (Objects.Count > 0) and (Found < LetterNum) do
    begin
      PO := Objects[0];
      Objects.Delete(0);
      Next := nil;
      if Objects.Count > 0 then
        Next := Objects[0];
      Dec(Found);
      MaxWidth := PO.Width;

      MinWidth := MinArea div PO.Height;
      if (TAntiCaptcha(FOwner).Script.GetParam('minimumLetterWidth') > 0) and (TAntiCaptcha(owner).Script.GetParam('minimumLetterWidth') > MinWidth) then
        MinWidth := TAntiCaptcha(FOwner).Script.GetParam('minimumLetterWidth');
      Splitter := 1;

      logger.Info(IntToStr(MaxWidth) + '/' + IntToStr(MinWidth));
      // ermittle die Vermutliche Buchstabenanzahl im Ersten captcha
      SplitNum := Min(Ceil(MaxWidth / (MinWidth / Splitter)), LetterNum - Found);
      while SplitNum < 2 do
      begin
        Inc(Splitter);
        SplitNum := Min(Ceil(MaxWidth / (MinWidth / Splitter)), LetterNum - Found);
      end;
      Logger.Info('l ' + IntToStr(SplitNum));

      while Found + SplitNum > LetterNum do
        Dec(SplitNum);

      Logger.Info('l ' + IntToStr(SplitNum));

      while (SplitNum > 2) and (Next <> nil) and (MaxWidth / SplitNum < Next.Width * 0.55) do
        Dec(SplitNum);

      Logger.Finer('teile erstes element ' + IntToStr(PO.Width) + ' : splitnum ' + IntToStr(SplitNum));

      if (Found + SplitNum - 1 > LetterNum) or (SplitNum < 2) then
      begin
        Logger.Severe('Richtige Letteranzahl konnte nicht ermittelt werden');
        Exit(nil);
      end;

      SplitObjects := PO.Split(SplitNum, TAntiCaptcha(FOwner).Script.GetParam('splitPixelObjectsOverlap'));

      Logger.Finer('Got splited: ' + IntToStr(SplitObjects.Count));
      // Fge die geteilen Objekte wieder dem Objektvektor hinzu.
      // Eventl
      // mssen sie nochmals geteil werden.

      for t := 0 to SplitNum - 1 do
      begin
        for s := 0 to Objects.Count - 1 do
          if (TPixelObject(SplitObjects[t]).Area > TPixelObject(Objects[s]).Area) then
          begin
            Objects.Insert(s, SplitObjects[t]);
            SplitObjects[t] := nil;
            Inc(Found);
            PerfectObjDetection := False;
            Logger.Finer('add split ' + IntToStr(Found));

            Break;
          end;

        if (SplitObjects[t] <> nil) then
        begin
          Objects.Add(SplitObjects[t]);
          SplitObjects[t] := nil;
          Inc(Found);
          PerfectObjDetection := False;
          Logger.Finer('add split ' + IntToStr(Found));
        end;
      end;

      Logger.Finer('Splitted ... Treffer: ' + IntToStr(Found));
    end;

    if (Found <> LetterNum) and (not TAntiCaptcha(FOwner).Script.GetParam('autoLetterNum')) then
    begin
      PerfectObjDetection := False;
      Logger.Severe('Richtige Letteranzahl konnte nicht ermittelt werden (#2)');
      Exit(nil);
    end;

    // entfernt berflssige Objekte
    j := Objects.Count - 1;
    while j >= Found do
    begin
      Objects.Delete(j);
      Dec(j);
    end;

    FPerfectObjectDetection := PerfectObjDetection;
  end;

  // Sortiert die Objekte nun endlich in der richtigen Reihenfolge (von
  // link nach rechts)
  Objects.Sort;

  Logger.Finer('Found ' + IntToStr(Objects.Count) + ' Elements');

  Result := Objects;
end;

procedure TCaptcha.SetPixel(APixel: TArrayofInteger);
var
  i, y, x: Integer;
  rgb: TArrayofInteger;
  Col: TArrayofDouble;
begin
  FPixel := APixel;
  i := 0;
  for y := 0 to GetHeight - 1 do
    for x := 0 to GetWidth - 1 do
    begin
      SetLength(Col, 4);

      if (FOwner as TAntiCaptcha).Script.ColorFormat = 0 then
      begin
        SetLength(rgb, 3);
        rgb[0] := GetRValue(APixel[i]);
        rgb[1] := GetGValue(APixel[i]);
        rgb[2] := GetBValue(APixel[i]);
        col := RGBToHSB(rgb);
      end
      else
        if (FOwner as TAntiCaptcha).Script.ColorFormat = 1 then
        begin
          col[0] := GetRValue(APixel[i]) / 255;
          col[1] := GetGValue(APixel[i]) / 255;
          col[2] := GetBValue(APixel[i]) / 255;
          col[3] := 0;
        end;

      FGrid[x][y] := Trunc(
                      (Col[(FOwner as TAntiCaptcha).Script.ColorComponents[0]] * 255) * 65536 +
                      (Col[(FOwner as TAntiCaptcha).Script.ColorComponents[1]] * 255) * 256   +
                      (Col[(FOwner as TAntiCaptcha).Script.ColorComponents[2]] * 255)
                    );
      FRGBGrid[x][y] := APixel[i];
      Inc(i);
    end;

end;

function TCaptcha.GetSimplified(Faktor: Double): TCaptcha;
var
  NewWidth, NewHeight, Avg, X, Y, NewX, NewY: Integer;
  NewGrid: T2DArrayofInteger;
begin
  if (Faktor = 1.0) or (Faktor = 0.0) then
    Exit(Self);

  NewWidth := Math.Ceil(GetWidth / Faktor);
  NewHeight := Math.Ceil(GetHeight / Faktor);
  Result := TCaptcha.Create(NewWidth, NewHeight);
  SetLength(NewGrid, NewWidth, NewHeight);
  Avg := GetAverage;

  for X := 0 to NewWidth - 1 do
    for Y := 0 to NewHeight - 1 do
      SetPixelValue(X, Y, NewGrid, GetMaxPixelValue);

  for X := 0 to GetWidth - 1 do
    for Y := 0 to GetHeight - 1 do
      if (IsElement(GetPixelValue(X, Y), Avg)) then
      begin
        NewX := Round(X / Faktor);
        NewY := Round(Y / Faktor);
        SetPixelValue(NewX, NewY, NewGrid, 0);
      end;

  Result.Grid := NewGrid;

  Result.PrintGrid;
end;

function TCaptcha.GetLetters(LetterNum: Integer; Contrast: Double; ObjectContrast: Double; MinArea: Integer): TArrayofLetter;
var
  Letters: TList<TPixelObject>;
  i: Integer;
  Obj: TPixelObject;
  L: TLetter;
begin
  Letters := GetBiggestObjects(LetterNum, MinArea, Contrast, ObjectContrast);
  if Letters = nil then
    Exit(nil);

  SetLength(FGaps, GetWidth + 1);
  for i := 0 to letters.Count - 1 do
  begin
    Obj := letters[i];

    L := TLetter(Obj.ToLetter);

    (FOwner as TAntiCaptcha).Script.ExecuteLetterPrepareCommands(L);

    SetLength(Result, Length(Result) + 1);
    Result[High(Result)] := L.GetSimplified(Double((FOwner as TAntiCaptcha).Script.GetParam('simplifyFaktor')));

    FGaps[TPixelObject(Letters[i]).GetLocation[0] + TPixelObject(Letters[i]).Width] := True;
  end;
end;

function TCaptcha.Reset: Boolean;
begin
  FSeparatedLetters := nil;
  FPrepared := False;
  if FTmpGrid = nil then
    Exit(False);

  FGrid := FTmpGrid;
  Result := True;
end;

function TCaptcha.GetLetters(LetterNum: Integer): TArrayofLetter;
begin
  if FSeparatedLetters <> nil then
    Exit(FSeparatedLetters);

  Result := GetLetters0(LetterNum);

  if Result = nil then
    Exit;

  // FEATURE UNIMPLEMENTED:
  // SPECIAL HANDLING

  FSeparatedLetters := Result;
end;

function TCaptcha.GetLetters(LetterNum: Integer; AGaps: TArrayofInteger): TArrayofLetter;
begin
  //
end;

function TCaptcha.GetLetters0(LetterNum: Integer): TArrayofLetter;
var
  Letters, LetterID: Integer;
  i: Integer;
  ClassName: string;
begin
  Result := nil;

  if LetterNum = 1 then
  begin
    Logger.Info('LetterNum = 1 --> Captcha to Letter');
    SetLength(Result, 1);
    Result[0] := TLetter(CreateLetter);
    Result[0].Grid := FGrid;
    Exit;
  end;

  if TAntiCaptcha(FOwner).Script.GetParam('useSpecialGetLetters') <> '' then
  begin
    ClassName := LowerCase(string(TAntiCaptcha(FOwner).Script.GetParam('useSpecialGetLetters')));

    if not SpecialClasses.ContainsKey(ClassName) then
    begin
      Logger.Warning('Special unbekannt: ' + ClassName);
      Exit;
    end;

    Exit(SpecialClasses[ClassName].GetLetters(Self));   // Error Insight sucks
  end;

  if Boolean(TAntiCaptcha(FOwner).Script.GetParam('useColorObjectDetection')) then
  begin
    Logger.Warning('FEATURE UNIMPLEMENTED: useColorObjectDetection');
    Exit;
  end;

  if Boolean(TAntiCaptcha(FOwner).Script.GetParam('useObjectDetection')) then
  begin
    Logger.Finer('Use Object Detection');

    Result := GetLetters(LetterNum,
                TAntiCaptcha(FOwner).Script.GetParam('ObjectColorContrast'),
                TAntiCaptcha(FOwner).Script.GetParam('ObjectDetectionContrast'),
                TAntiCaptcha(FOwner).Script.GetParam('MinimumObjectArea') );

    if Result <> nil then
      Exit
    else
      Logger.Severe('Object detection failed. Try alternative Methods');
  end;

  if Boolean(TAntiCaptcha(FOwner).Script.GetParam('cancelIfObjectDetectionFailed')) then
    Exit(nil);

  if (not TAntiCaptcha(FOwner).Script.GetParam('UseAverageGapDetection')) and
     (not TAntiCaptcha(FOwner).Script.GetParam('UsePeakGapdetection')) and
     (TAntiCaptcha(FOwner).Script.Gaps <> nil) then
  begin
    Logger.Finer('Use predefined Gaps');
    Exit(GetLetters(LetterNum, TAntiCaptcha(FOwner).Script.Gaps));
  end;

  Logger.Finer('Use Line Detection');

  SetLength(FGaps, GetWidth + 1);
  SetLength(Result, LetterNum);

  FLastLetterX := 0;
  Letters := 0;

  for LetterID := 0 to LetterNum - 1 do
  begin
    Result[LetterID] := GetNextLetter(LetterID);

    if (Result[LetterID] = nil) and (not TAntiCaptcha(FOwner).Script.GetParam('autoLetterNum')) then
    begin
      if (TAntiCaptcha(FOwner).Script.Gaps <> nil) then
        Exit(GetLetters(LetterNum, TAntiCaptcha(FOwner).Script.Gaps))
      else
      begin
        for i := 0 to LetterID - 1 do
          Result[i].Free;

        Logger.Fine('LetterID #' + IntToStr(LetterID) + ' = nil!');

        Exit(nil);
      end;
    end
    else
      if Result[LetterID] <> nil then
      begin
        Inc(Letters);
        TAntiCaptcha(FOwner).Script.ExecuteLetterPrepareCommands(Result[LetterID]);

        Result[LetterID] := Result[LetterID].GetSimplified(TAntiCaptcha(FOwner).Script.GetParam('simplifyFaktor'));
      end;
  end;

  if (TAntiCaptcha(FOwner).Script.GetParam('autoLetterNum')) and (Letters < Length(Result)) then
    Result := Copy(Result, 0, Letters);
end;

function TCaptcha.GetNextLetter(LetterID: Integer): TLetter;
var
  LetterGrid: T2DArrayofInteger;
  RowAverage, RowPeak: TArrayofInteger;
  i, Average, X, Y, NoGapCount, Count, PixelValue, Line: Integer;
  LastOverPeak, IsGap, IsAverageGap, IsOverPeak, IsPeakGap: Boolean;
  Loc: TStatic01ArrayofInt;
begin
  Result := TLetter(CreateLetter);

  SetLength(LetterGrid, GetWidth, GetHeight);
  SetLength(RowAverage, GetWidth);
  SetLength(RowPeak, GetWidth);

  for i := 0 to High(RowAverage) do
  begin
    RowAverage[i] := 0;
    RowPeak[i] := MAXINT;
  end;

  Average := GetAverage;
  NoGapCount := 0;
  LastOverPeak := False;

  for x := FLastLetterX to GetWidth - 1 do
  begin
    Count := 0;

    for Y := 0 to GetHeight - 1 do
    begin
      for Line := 0 to TAntiCaptcha(FOwner).Script.GetParam('GapWidthPeak') - 1 do
        if GetWidth > X + Line then
        begin
          PixelValue := GetPixelValue(x + line, y);
          if PixelValue < RowPeak[x] then
            RowPeak[x] := PixelValue;
        end;

      for Line := 0 to TAntiCaptcha(FOwner).Script.GetParam('GapWidthAverage') - 1 do
        if GetWidth > X + Line then
        begin
          RowAverage[x] := MixColors(RowAverage[x], GetPixelValue(X + Line, Y), Count, 1);
          Inc(Count);
        end;

      LetterGrid[x][y] := GetPixelValue(X, Y);
    end;

    if Boolean(TAntiCaptcha(FOwner).Script.GetParam('GapAndAverageLogic')) then
    begin
      IsAverageGap := (rowAverage[x] > average * TAntiCaptcha(owner).Script.GetParam('GapDetectionAverageContrast')) or (not TAntiCaptcha(FOwner).Script.GetParam('UseAverageGapDetection'));

      IsOverPeak := RowPeak[x] < Average * TAntiCaptcha(FOwner).Script.GetParam('GapDetectionPeakContrast');
      IsPeakGap := LastOverPeak and (not IsOverPeak) or (not TAntiCaptcha(FOwner).Script.GetParam('UsePeakGapdetection'));

      IsGap := IsAverageGap and IsPeakGap;
    end
    else
    begin
      IsAverageGap := (RowAverage[x] > Average * TAntiCaptcha(FOwner).Script.GetParam('GapDetectionAverageContrast')) and (TAntiCaptcha(owner).Script.getParam('UseAverageGapDetection'));

      IsOverPeak := RowPeak[x] < Average * TAntiCaptcha(FOwner).Script.GetParam('GapDetectionPeakContrast');
      IsPeakGap := LastOverPeak and (not IsOverPeak) or (not TAntiCaptcha(FOwner).Script.GetParam('UsePeakGapdetection'));

      IsGap := IsAverageGap or IsPeakGap;
    end;
    LastOverPeak := IsOverPeak;

    if (IsGap) and (NoGapCount > TAntiCaptcha(FOwner).Script.GetParam('MinimumLetterWidth')) then
      Break
    else
    if (rowAverage[x] < average * TAntiCaptcha(FOwner).Script.GetParam('GapDetectionAverageContrast')) then
      Inc(NoGapCount);
  end;

  Result.Grid := LetterGrid;
  Loc[0] := 0;
  Loc[1] := 0;
  Result.Location := Loc;

  if not Result.Trim(FLastLetterX, X) then
  begin
    Logger.Severe(Format('Trimming failed (%d, %d)', [FLastLetterX, X]));
    FreeAndNil(Result);
    Exit;
  end;

  if not Result.Clean then
  begin
    FreeAndNil(Result);
    Exit;
  end;

  FLastLetterX := X;

  FGaps[Min(FLastLetterX, GetWidth)] := True;
end;

procedure TCaptcha.MergeMultiPartObjects(var Objects: TList<TPixelObject>);
begin
  if (TAntiCaptcha(owner).Script.getParam('multiplePartMergeMinSize') <= 0) then
    Exit;

  // xxx
  raise Exception.Create('FEATURE UNIMPLEMENTED: MergeMultiPartObjecte!');
end;

procedure TCaptcha.CleanByRGBDistance(Mode, Tollerance: Integer);
var
  NewGrid: T2DArrayofInteger;
  x, y, Val: Integer;
begin
  SetLength(NewGrid, GetWidth, GetHeight);

  for x := 0 to GetWidth - 1 do
    for y := 0 to GetHeight - 1 do
    begin
      Val := GetRGBDistance(GetPixelValue(x, y));

      if Mode = 1 then
      begin
        if (Val > Tollerance) then
          SetPixelValue(x, y, NewGrid, GetMaxPixelValue)
        else
          NewGrid[x][y] := FGrid[x][y];
      end
      else
        if (Val < tollerance) then
          SetPixelValue(x, y, NewGrid, GetMaxPixelValue)
        else
          NewGrid[x][y] := FGrid[x][y];
    end;

  FGrid := NewGrid;
end;

end.
