(*
     Catcher! OCR Module  -  Port of JAntiCaptcha
     Copyright (C) 2010  CandiSoft

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

unit Letter;

interface

uses PixelGrid, Math, UTILITIES, SysUtils, CSTools, untLogger;

type
  TLetter = class(TPixelGrid)
  private
    FDecodedValue, FSourceHash: string;
    FID, FGoodDetections, FElementPixel, FBadDetections, FAngle: Integer;
    FDetected: TObject;    // TLetterComperator

    procedure FollowPixelObject(X, Y: Integer; Map: T2DArrayofInteger; BMap: TBooleanMap);
    function GetFirstAndLastLinePixels: Integer;
    function GetObjectsNum(Map: T2DArrayofInteger): Integer;

    function GetArea: Integer;
  public
    constructor Create;

    function Align(ObjectContrast: Double): TLetter; overload;
    function Align(ObjectContrast: Double; AngleA, AngleB: Integer): TLetter; overload;
    function Align(AngleA, AngleB: Integer): TLetter; overload;
    function BetterAlign(AngleA, AngleB: Integer): TLetter;
    procedure Colorize(Color: Integer);
    procedure FillWithObject(P: TObject; Color: Integer);
    function GetElementPixel: Integer;
    function GetLinedLetter: TLetter;
    function GetLocalMap(AGrid: T2DArrayofInteger; X, Y: Integer): T2DArrayofInteger;
    function GetMassLetter: TLetter;
    function GetPixelString: string;
    function GetSimplified(faktor: Double): TLetter;
    procedure MarkBad;
    procedure MarkGood;
    procedure ResizeToHeight(NewHeight: Integer); overload;
    procedure ResizeToHeight(NewHeight: Integer; D: Double); overload;
    procedure SetElementPixel(ElementPixel: Integer);
    function SetTextGrid(Content: string): Boolean;
    function ToPixelObject(ObjectContrast: Double): TObject;
    function ToString: string; override;
    function Trim(Left, Right: Integer): Boolean;
    function Turn(Angle: Double): TLetter;

    property Area: Integer read GetArea;
    property Angle: Integer read FAngle write FAngle;
    property BadDetections: Integer read FBadDetections write FBadDetections;
    property DecodedValue: string read FDecodedValue write FDecodedValue;
    property Detected: TObject read FDetected write FDetected;
    property GoodDetections: Integer read FGoodDetections write FGoodDetections;
    property ID: Integer read FID write FID;
    property SourceHash: string read FSourceHash write FSourceHash;
  end;

implementation

uses PixelObject, AntiCaptcha, LetterComperator;

constructor TLetter.Create;
begin
  SetLength(FGrid, 0, 0);
  FDetected := nil;
end;

function TLetter.Align(ObjectContrast: Double): TLetter;
var
  Obj, Aligned: TPixelObject;
  NewLetter: TLetter;
begin
  Obj := TPixelObject(ToPixelObject(ObjectContrast));
  try
    Aligned := Obj.Align;
    NewLetter := TLetter(Aligned.ToLetter);
  finally
    Obj.Free;
    Aligned.Free;
  end;


  FGrid := NewLetter.Grid;
  Result := Self;
end;

function TLetter.Align(ObjectContrast: Double; AngleA, AngleB: Integer): TLetter;
var
  Obj, Aligned: TPixelObject;
  NewLetter: TLetter;
begin
  Obj := TPixelObject(ToPixelObject(ObjectContrast));
  try
    Aligned := Obj.Align(AngleA, AngleB);
    NewLetter := TLetter(Aligned.ToLetter);
  finally
    Obj.Free;
    Aligned.Free;
  end;


  FGrid := NewLetter.Grid;
  Result := Self;
end;

function TLetter.Align(AngleA: Integer; AngleB: Integer): TLetter;
var
  Tmp, Accuracy: Integer;
  BestValue: Double;
  LTmp: TLetter;
begin
  if AngleB < AngleA then
  begin
    Tmp := AngleB;
    AngleB := AngleA;
    AngleA := Tmp;
  end;

  Accuracy := (Owner as TAntiCaptcha).Script.GetParam('AlignAngleSteps');
  BestValue := MaxDouble;

  Result := nil;

  FAngle := AngleA;
  while FAngle < AngleB do
  begin
    if FAngle < 0 then
      lTmp := Turn(360 + FAngle)
    else
      lTmp := Turn(FAngle);

    if LTmp.GetWidth / LTmp.GetHeight < BestValue then
    begin
      BestValue := LTmp.GetWidth / LTmp.GetHeight;
      Result := LTmp;
    end;

    FAngle := FAngle + Accuracy;
  end;
end;

function TLetter.BetterAlign(AngleA: Integer; AngleB: Integer): TLetter;
var
  Tmp, Accuracy, BestPix, Angle, Pix: Integer;
  LTmp: TLetter;
begin
  if AngleB < AngleA then
  begin
    Tmp := AngleB;
    AngleB := AngleA;
    AngleA := Tmp;
  end;

  Accuracy := (Owner as TAntiCaptcha).Script.GetParam('AlignAngleSteps');
  BestPix := GetFirstAndLastLinePixels;

  Result := nil;

  Angle := AngleA;
  while Angle < AngleB do
  begin
    if Angle < 0 then
      LTmp := Turn(360 + Angle)
    else
      LTmp := Turn(Angle);

    Pix := LTmp.GetFirstAndLastLinePixels;
    if Pix > BestPix then
    begin
      BestPix := Pix;
      Result := LTmp;
    end;

    Inc(Angle, Accuracy);
  end;
end;

procedure TLetter.Colorize(Color: Integer);
var
  X, Y: Integer;
begin
  for X := 0 to GetWidth - 1 do
    for Y := 0 to GetHeight - 1 do
       if FGrid[x][y] < 40 then
         FGrid[x][y] := Color;
end;

procedure TLetter.FillWithObject(P: TObject; Color: Integer);
var
  i: Integer;
  Point: TArrayofInteger;
begin
  for i := 0 to (P as TPixelObject).Size - 1 do
  begin
    Point := (P as TPixelObject).ElementAt(i);

    if Color < 0 then
      FGrid[Point[0]][Point[1]] := Point[2]
    else
      FGrid[Point[0]][Point[1]] := Color;
  end;
end;

procedure TLetter.FollowPixelObject(X: Integer; Y: Integer; Map: T2DArrayofInteger; BMap: TBooleanMap);
var
  PosX, PosY: Integer;
begin
  if BMap[X][Y] then
    Exit;

  BMap[X][Y] := True;

  // links oben
  PosX := -1;
  PosY := -1;

  try
    if (Map[X + PosX][Y + PosY] = 0) then
      FollowPixelObject(X + PosX, Y + PosY, Map, BMap);
  except end;

  // links
  PosX := -1;
  PosY := 0;

  try
    if (Map[X + PosX][Y + PosY] = 0) then
      FollowPixelObject(X + PosX, Y + PosY, Map, BMap);
  except end;

  // links unten
  PosX := -1;
  PosY := 1;

  try
    if (Map[X + PosX][Y + PosY] = 0) then
      FollowPixelObject(X + PosX, Y + PosY, Map, BMap);
  except end;

  // unten
  PosX := 0;
  PosY := 1;

  try
    if (Map[X + PosX][Y + PosY] = 0) then
      FollowPixelObject(X + PosX, Y + PosY, Map, BMap);
  except end;

  // rechts unten
  PosX := 1;
  PosY := 1;

  try
    if (Map[X + PosX][Y + PosY] = 0) then
      FollowPixelObject(X + PosX, Y + PosY, Map, BMap);
  except end;

  // rechts
  PosX := 1;
  PosY := 0;

  try
    if (Map[X + PosX][Y + PosY] = 0) then
      FollowPixelObject(X + PosX, Y + PosY, Map, BMap);
  except end;

  // rechts oben
  PosX := 1;
  PosY := -1;

  try
    if (Map[X + PosX][Y + PosY] = 0) then
      FollowPixelObject(X + PosX, Y + PosY, Map, BMap);
  except end;

  // oben
  PosX := 0;
  PosY := -1;

  try
    if (Map[X + PosX][Y + PosY] = 0) then
      FollowPixelObject(X + PosX, Y + PosY, Map, BMap);
  except end;
end;

function TLetter.GetArea: Integer;
begin
  Result := GetWidth * GetHeight;
end;

function TLetter.GetElementPixel: Integer;
var
  X, Y: Integer;
begin
  if FElementPixel > 0 then
    Exit(FElementPixel);

  FElementPixel := 0;
  for X := 0 to GetWidth - 1 do
    for Y := 0 to GetHeight - 1 do
      if FGrid[X][Y] = 0 then
        Inc(FElementPixel);

  Result := FElementPixel;
end;

function TLetter.GetFirstAndLastLinePixels: Integer;
var
  c, Avg, X, Y, c2: Integer;
begin
  c := 0;
  Avg := GetAverage;

  for Y := 0 to GetHeight - 1 do
  begin
    for X := 0 to GetWidth - 1 do
      if IsElement(GetPixelValue(X, Y), Avg) then
        Inc(c);

    if c > 0 then
      Break;
  end;

  c2 := 0;

  for Y := GetHeight - 1 downto 1 do
  begin
    for X := 0 to GetWidth - 1 do
      if IsElement(GetPixelValue(X, Y), Avg) then
        Inc(c2);

    if c2 > 0 then
     	Break;
  end;

  Result := c + c2;
end;

function TLetter.GetLinedLetter: TLetter;
var
  FirstChange, Changed, i, AX, AY, A, B: Integer;
  PO: TPixelObject;
  Border, Map: T2DArrayofInteger;
begin
  Result := TLetter.Create;
  Result.Owner := FOwner;

  FirstChange := 0;
  while True do
  begin
    Changed := 0;
    PO := TPixelObject(ToPixelObject(0.85));
    Border := PO.GetBorderVector(Self);
    try


      for i := 0 to High(Border) do
      begin
        AX := Border[i][0];
        AY := Border[i][1];
        Logger.Info(Format('%d/%d', [ax, ay]));
        Map := GetLocalMap(FGrid, AX, AY);

        A := GetObjectsNum(Map);
        Map[1][1] := $FF000;

        B := GetObjectsNum(Map);
        Logger.Info(IntToStr(A) + ' --->> ' + IntToStr(B));

        if A = B then
        begin
          Inc(Changed);
          SetPixelValue(AX, AY, $FF0000);
        end
        else
        begin
          Logger.Info(IntToStr(Map[0][0]) + '-' + IntToStr(Map[1][0]) + '-' + IntToStr(Map[2][0]));
          Logger.Info(IntToStr(Map[0][1]) + '-' + IntToStr(Map[1][1]) + '-' + IntToStr(Map[2][1]));
          Logger.Info(IntToStr(Map[0][2]) + '-' + IntToStr(Map[1][2]) + '-' + IntToStr(Map[2][2]));
          SetPixelValue(AX, AY, 0);
        end;

        Map := nil;
      end;

      // frmGUI.ShowImage(GetImage(5));
      Logger.Info('changed ' + IntToStr(Changed));
      if FirstChange = 0 then
        FirstChange := Changed;

    finally
      PO.Free;
      Border := nil;
    end;

    if Changed * 20 <= FirstChange then
      Break;
  end;

  Result.Grid := FGrid;
end;

function TLetter.GetLocalMap(AGrid: T2DArrayofInteger; X, Y: Integer): T2DArrayofInteger;
var
  XX, YY: Integer;
begin
  SetLength(Result, 3, 3);

  for XX := -1 to 1 do
    for YY := -1 to 1 do
      try
        Result[XX + 1][YY + 1] := AGrid[X + XX][Y + YY];
      except
        Result[XX + 1][YY + 1] := -1;
      end;
end;

function TLetter.GetMassLetter: TLetter;
var
  NewGrid, FilterGrid, Map: T2DArrayofInteger;
  Grd: array of array of Int64;
  PO: TPixelObject;
  Max, Min, Mass: Int64;
  X, Y, i, XMax, YMax, Value, XX, YY: Integer;
begin
  Result := TLetter.Create;
  Result.Owner := FOwner;

  SetLength(NewGrid, GetWidth, GetHeight);
  SetLength(FilterGrid, GetWidth, GetHeight);
  SetLength(Grd, GetWidth, GetHeight);
  PO := TPixelObject(ToPixelObject(0.85));
  try
    Max := 0;
    Min := 9223372036854775807;    // MaxInt64
    for X := 0 to GetWidth - 1 do
      for Y := 0 to GetHeight - 1 do
      begin
        Mass := PO.GetMassValue(X, Y);
        Max := Math.Max(Max, Mass);
        Min := Math.Min(Min, Mass);

        Grd[X][Y] := Mass;
      end;
  finally
    PO.Free;
  end;


  for X := 0 to GetWidth - 1 do
    for Y := 0 to GetHeight - 1 do
    begin
      FilterGrid[X][Y] := 0;
      NewGrid[X][Y] := Trunc(GetMaxPixelValue * ((Grd[X][Y] - Min) / (Max - Min)));
    end;

  Grd := nil;

  X := 0;
  Y := 0;
  SetLength(Map, 3, 3);
  i := 1000000;
  while True do
  begin
    Map := GetLocalMap(NewGrid, X, Y);
    // Logger.Info(IntToStr(x) + ' - ' + IntToStr(y));
    try
      FilterGrid[X][Y] := NewGrid[X][Y];

      XMax := -2;
      YMax := -2;
      Value := -1;
      for XX := -1 to 1 do
        for YY := -1 to 1 do
          if (YY <> 0) and (XX <> 0) then
            if (Map[XX + 1][YY + 1] >= Value) and (Map[XX + 1][YY + 1] > 0) and
               (FilterGrid[X + XX][Y + YY] = 0) then
            begin
              Value := Map[XX + 1][YY + 1];
              XMax := XX;
              YMax := YY;
            end;

      if XMax < -1 then
        XMax := 1;
      if YMax < -1 then
        YMax := 1;

      Inc(X, XMax);
      Inc(Y, YMax);

      Dec(i);
      if i = 0 then
        Break;
    except
      Break;
    end;

    Map := nil;
  end;

  FilterGrid := nil;

  Result.Grid := NewGrid;
end;

function TLetter.GetObjectsNum(Map: T2DArrayofInteger): Integer;
var
  BMap: TBooleanMap;
  X, Y: Integer;
begin
  Result := 0;

  // Initialisieren
  for X := 0 to 2 do
    for Y := 0 to 2 do
      BMap[X][Y] := False;

  for X := 0 to 2 do
    for Y := 0 to 2 do
      if (not BMap[X][Y]) and (Map[X][Y] = 0) then
      begin
        FollowPixelObject(X, Y, Map, BMap);
        Inc(Result);
      end;
end;

function TLetter.GetPixelString: string;
var
  X, Y: Integer;
begin
  Result := '';

  for X := 0 to GetWidth - 1 do
  begin
    for Y := 0 to GetHeight - 1 do
      Result := Result + IntToStr(GetPixelValue(X, Y) div GetMaxPixelValue);

    Result := Result + '|';
  end;

  Result := Copy(Result, 1, Length(Result) - 1);
end;

function TLetter.GetSimplified(Faktor: Double): TLetter;
var
  NewWidth, NewHeight, Avg, X, Y, NewX, NewY: Integer;
  NewGrid: T2DArrayofInteger;
begin
  if (Faktor = 1.0) or (Faktor = 0.0) then
    Exit(Self);

  NewWidth := Ceil(GetWidth / Faktor);
  NewHeight := Ceil(GetHeight / Faktor);
  Result := TLetter.Create;
  Result.Owner := FOwner;
  Avg := GetAverage;

  Logger.Info('GetSimplified: ' + IntToStr(NewWidth) + ' - ' + IntToStr(NewHeight));

  SetLength(NewGrid, NewWidth, NewHeight);

  for X := 0 to NewWidth - 1 do
    for Y := 0 to NewHeight - 1 do
      SetPixelValue(X, Y, NewGrid, GetMaxPixelValue);

  FElementPixel := 0;

  for X := 0 to GetWidth - 1 do
    for Y := 0 to GetHeight - 1 do
    begin
      if IsElement(GetPixelValue(X, Y), Avg) then
      begin
        NewX := Floor(X / Faktor);
        NewY := Floor(Y / Faktor);
        SetPixelValue(NewX, NewY, NewGrid, 0);
      end;
    end;

  Result.Grid := NewGrid;
  Result.Clean;
end;

procedure TLetter.MarkBad;
begin
  Inc(FBadDetections);
  Logger.Warning('Bad detection : (' + ToString + ')');
end;

procedure TLetter.MarkGood;
begin
  Inc(FGoodDetections);
  Logger.Warning('GOOD detection : (' + ToString + ')');
end;

procedure TLetter.ResizeToHeight(NewHeight: Integer);
var
  Faktor: Double;
  NewWidth, ElementPixel, X, Y, V: Integer;
  NewGrid: T2DArrayofInteger;
begin
  Faktor := NewHeight / GetHeight;
  NewWidth := Ceil(GetWidth * Faktor);

  SetLength(NewGrid, NewWidth, NewHeight);
  ElementPixel := 0;

  for X := 0 to NewWidth - 1 do
    for Y := 0 to NewHeight - 1 do
    begin
      V := FGrid[Floor(X / Faktor)][Floor(Y / Faktor)];
      NewGrid[X][Y] := V;
      if NewGrid[X][Y] = 0 then
        Inc(ElementPixel);
    end;

  SetElementPixel(ElementPixel);
  FGrid := NewGrid;
end;

procedure TLetter.ResizeToHeight(NewHeight: Integer; D: Double);
var
  Faktor: Double;
begin
  Faktor := NewHeight / GetHeight;
  if Abs(Faktor - 1) < D then
    ResizeToHeight(NewHeight);
end;

procedure TLetter.SetElementPixel(ElementPixel: Integer);
begin
  FElementPixel := ElementPixel;
end;

function TLetter.SetTextGrid(Content: string): Boolean;
var
  Code: TArrayofstring;
  Width, ElementPixel, X, Y: Integer;
  Line: string;
begin
  Code := Explode('|', Content);
  FGrid := nil;
  ElementPixel := 0;
  Result := True;

  for Y := 0 to High(Code) do
  begin
    Line := Code[Y];
    Width := Length(Line);

    if FGrid = nil then
    begin
      SetLength(FGrid, Width, Length(Code));
      if (Width < 1) or (Length(Code) < 2) then      // xxx Width beobachten, war vorher < 2
        Exit(False);
    end;

    for X := 0 to Width - 1 do
    begin
      try
        FGrid[X][Y] := StrToInt(Line[X + 1]) * GetMaxPixelValue;   // + 1 weil in Delphi strings mit Index 1 beginnen!
      except
        Logger.Severe('Parsing TextGrid failed: ' + Content);
        Result := False;
      end;

      if not Result then
        Exit;

      if FGrid[X][Y] = 0 then
        Inc(ElementPixel);
    end;
  end;

  SetElementPixel(ElementPixel);
end;

function TLetter.ToPixelObject(ObjectContrast: Double): TObject;
var
  PObject: TPixelObject;
  X, Y: Integer;
begin
  PObject := TPixelObject.Create(Self);
  PObject.WhiteContrast := ObjectContrast;

  for X := 0 to GetWidth - 1 do
    for Y := 0 to GetHeight - 1 do
      if GetPixelValue(X, Y) < (GetMaxPixelValue * ObjectContrast) then
        PObject.Add(X, Y, GetPixelValue(X, Y));

  Result := PObject;
end;

function TLetter.ToString: string;
begin
  Result := FDecodedValue + ' [' + FSourcehash +'][' + IntToStr(FGoodDetections) + '/' + IntToStr(FBadDetections) + ']';
end;

function TLetter.Trim(Left: Integer; Right: Integer): Boolean;
var
  width, x: Integer;
  tmp: T2DArrayofInteger;
  loc: TStatic01ArrayofInt;
begin
  width := right - left;
  SetLength(tmp, width, getHeight);
  loc[0] := left;
  loc[1] := 0;
  FLocation := loc;
  if (getWidth < right) then
  begin
    Logger.Severe('Letter dim: ' + IntToStr(getWidth) + ' - ' + IntToStr(getHeight) + '. Cannot trim to ' + IntToStr(left) + '-' + IntToStr(right));
    Exit(False);
  end;

  for x := 0 to width - 1 do
    tmp[x] := grid[x + left];

  grid := tmp;
  Result := True;
end;

function TLetter.Turn(Angle: Double): TLetter;
var
  NewWidth, NewHeight, left, top, ElementPixel, x, y: Integer;
  NewGrid: T2DArrayofInteger;
  n: TArrayofInteger;
begin
  if Angle = 0.0 then
    Exit(Self);

  while Angle < 0 do
    Angle := Angle + 360;

  Angle := Angle / 180;
  Result := TLetter(CreateLetter);

  NewWidth := Trunc(Abs(Cos(Angle * Pi) * GetWidth) + Abs(Sin(Angle * Pi) * GetHeight));
  NewHeight := Trunc(Abs(Sin(Angle * Pi) * GetWidth) + Abs(Cos(Angle * Pi) * GetHeight));

 // logger.info(IntToStr(getWidth)+'/'+IntToStr(getHeight)+'--> '+IntToStr(newWidth)+'/'+IntToStr(newHeight)
 //        +'('+FloatToStr(angle)+'/'+FloatToStr((angle*180))+'/');

  left := (NewWidth - GetWidth) div 2;
  top := (NewHeight - GetHeight) div 2;
  ElementPixel := 0;
  SetLength(NewGrid, NewWidth, NewHeight);
  for x := 0 to NewWidth - 1 do
    for y := 0 to NewHeight - 1 do
    begin
      n := TurnCoordinates(x - left, y - top, GetWidth div 2, GetHeight div 2, -(Angle * 180));
      if (n[0] < 0) or (n[0] >= GetWidth) or (n[1] < 0) or (n[1] >= GetHeight) then
      begin
        NewGrid[x][y] := (Owner as TAntiCaptcha).Script.ColorFaktor - 1;
        if NewGrid[x][y] = 0 then
          Inc(ElementPixel);

        Continue;
      end;

      NewGrid[x][y] := FGrid[n[0]][n[1]];
      if NewGrid[x][y] = 0 then
        Inc(ElementPixel);
    end;

  Result.Grid := NewGrid;
  Result.SetElementPixel(ElementPixel);
  Result.Clean;
  Result.DecodedValue := FDecodedValue;
  Result.Angle := Trunc(Angle * 180.0);
end;

end.
